# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from .callbacks import Callback
from .utils import PyTorchTypes


class _CallbackInterface:
    """
    An interface of all the methods every callback must implement for dynamically calling the methods from the callbacks
    handler.
    """

    ON_SETUP = "on_setup"
    ON_RUN_BEGIN = "on_run_begin"
    ON_RUN_END = "on_run_end"
    ON_EPOCH_BEGIN = "on_epoch_begin"
    ON_EPOCH_END = "on_epoch_end"
    ON_TRAIN_BEGIN = "on_train_begin"
    ON_TRAIN_END = "on_train_end"
    ON_VALIDATION_BEGIN = "on_validation_begin"
    ON_VALIDATION_END = "on_validation_end"
    ON_TRAIN_BATCH_BEGIN = "on_train_batch_begin"
    ON_TRAIN_BATCH_END = "on_train_batch_end"
    ON_VALIDATION_BATCH_BEGIN = "on_validation_batch_begin"
    ON_VALIDATION_BATCH_END = "on_validation_batch_end"
    ON_INFERENCE_BEGIN = "on_inference_begin"
    ON_INFERENCE_END = "on_inference_end"
    ON_TRAIN_LOSS_BEGIN = "on_train_loss_begin"
    ON_TRAIN_LOSS_END = "on_train_loss_end"
    ON_VALIDATION_LOSS_BEGIN = "on_validation_loss_begin"
    ON_VALIDATION_LOSS_END = "on_validation_loss_end"
    ON_TRAIN_METRICS_BEGIN = "on_train_metrics_begin"
    ON_TRAIN_METRICS_END = "on_train_metrics_end"
    ON_VALIDATION_METRICS_BEGIN = "on_validation_metrics_begin"
    ON_VALIDATION_METRICS_END = "on_validation_metrics_end"
    ON_BACKWARD_BEGIN = "on_backward_begin"
    ON_BACKWARD_END = "on_backward_end"
    ON_OPTIMIZER_STEP_BEGIN = "on_optimizer_step_begin"
    ON_OPTIMIZER_STEP_END = "on_optimizer_step_end"
    ON_SCHEDULER_STEP_BEGIN = "on_scheduler_step_begin"
    ON_SCHEDULER_STEP_END = "on_scheduler_step_end"
    ON_CALL_CHECK = "on_call_check"


class CallbacksHandler:
    """
    A class for handling multiple callbacks during a run.
    """

    def __init__(self, callbacks: list[Union[Callback, tuple[str, Callback]]]):
        """
        Initialize the callbacks handler with the given callbacks he will handle. The callbacks can be passed as their
        initialized instances or as a tuple where [0] is a name that will be attached to him and [1] will be the
        initialized callback instance. Notice that if a callback was given without a name, his name to refer will be
        his class name.

        :param callbacks: The callbacks to handle.

        :raise KeyError: In case of duplicating callbacks names.
        """
        # Initialize the callbacks dictionary:
        self._callbacks = {}  # type: Dict[str, Callback]

        # Add the given callback to the dictionary:
        for callback in callbacks:
            if isinstance(callback, tuple):
                # Add by given name:
                if callback[0] in self._callbacks:
                    raise KeyError(
                        f"A callback with the name '{callback[0]}' is already exist in the callbacks handler."
                    )
                self._callbacks[callback[0]] = callback[1]
            else:
                # Add by class name:
                if callback.__class__.__name__ in self._callbacks:
                    raise KeyError(
                        f"A callback with the name '{callback.__class__.__name__}' is already exist in the callbacks "
                        f"handler."
                    )
                self._callbacks[callback.__class__.__name__] = callback

    @property
    def callbacks(self) -> dict[str, Callback]:
        """
        Get the callbacks dictionary handled by this handler.

        :return: The callbacks dictionary.
        """
        return self._callbacks

    def on_setup(
        self,
        model: Module,
        training_set: DataLoader,
        validation_set: DataLoader,
        loss_function: Module,
        optimizer: Optimizer,
        metric_functions: list[PyTorchTypes.MetricFunctionType],
        scheduler,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_setup' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param model:            The model to pass to the method.
        :param training_set:     The training set to pass to the method.
        :param validation_set:   The validation set to pass to the method.
        :param loss_function:    The loss function to pass to the method.
        :param optimizer:        The optimizer to pass to the method.
        :param metric_functions: The metric functions to pass to the method.
        :param scheduler:        The scheduler to pass to the method.
        :param callbacks:        The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_SETUP,
            callbacks=self._parse_names(names=callbacks),
            model=model,
            training_set=training_set,
            validation_set=validation_set,
            loss_function=loss_function,
            optimizer=optimizer,
            metric_functions=metric_functions,
            scheduler=scheduler,
        )

    def on_run_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_run_begin' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_RUN_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_run_end(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_run_end' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_RUN_END,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_epoch_begin(self, epoch: int, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_epoch_begin' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param epoch:     The current epoch iteration of when this method is called.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_EPOCH_BEGIN,
            callbacks=self._parse_names(names=callbacks),
            epoch=epoch,
        )

    def on_epoch_end(self, epoch: int, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_epoch_end' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param epoch:     The current epoch iteration of when this method is called.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_EPOCH_END,
            callbacks=self._parse_names(names=callbacks),
            epoch=epoch,
        )

    def on_train_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_train_begin' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_train_end(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_train_end' method of every callback in the callbacks list. If the list is 'None' (not given), all
        callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_END,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_validation_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_validation_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_validation_end(
        self,
        loss_value: PyTorchTypes.MetricValueType,
        metric_values: list[float],
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_validation_end' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param loss_value:    The loss summary of this validation.
        :param metric_values: The metrics summaries of this validation.
        :param callbacks:     The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_END,
            callbacks=self._parse_names(names=callbacks),
            loss_value=loss_value,
            metric_values=metric_values,
        )

    def on_train_batch_begin(
        self, batch: int, x, y_true: Tensor, callbacks: Optional[list[str]] = None
    ) -> bool:
        """
        Call the 'on_train_batch_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param batch:     The current batch iteration of when this method is called.
        :param x:         The input part of the current batch.
        :param y_true:    The true value part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_BATCH_BEGIN,
            callbacks=self._parse_names(names=callbacks),
            batch=batch,
            x=x,
            y_true=y_true,
        )

    def on_train_batch_end(
        self,
        batch: int,
        x,
        y_pred: Tensor,
        y_true: Tensor,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_train_batch_end' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param batch:     The current batch iteration of when this method is called.
        :param x:         The input part of the current batch.
        :param y_pred:    The prediction (output) of the model for this batch's input ('x').
        :param y_true:    The true value part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_BATCH_END,
            callbacks=self._parse_names(names=callbacks),
            batch=batch,
            x=x,
            y_pred=y_pred,
            y_true=y_true,
        )

    def on_validation_batch_begin(
        self, batch: int, x, y_true: Tensor, callbacks: Optional[list[str]] = None
    ) -> bool:
        """
        Call the 'on_validation_batch_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param batch:     The current batch iteration of when this method is called.
        :param x:         The input part of the current batch.
        :param y_true:    The true value part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_BATCH_BEGIN,
            callbacks=self._parse_names(names=callbacks),
            batch=batch,
            x=x,
            y_true=y_true,
        )

    def on_validation_batch_end(
        self,
        batch: int,
        x,
        y_pred: Tensor,
        y_true: Tensor,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_validation_batch_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param batch:     The current batch iteration of when this method is called.
        :param x:         The input part of the current batch.
        :param y_pred:    The prediction (output) of the model for this batch's input ('x').
        :param y_true:    The true value part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_BATCH_END,
            callbacks=self._parse_names(names=callbacks),
            batch=batch,
            x=x,
            y_pred=y_pred,
            y_true=y_true,
        )

    def on_inference_begin(
        self,
        x,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_inference_begin' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param x:         The input part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_INFERENCE_BEGIN,
            callbacks=self._parse_names(names=callbacks),
            x=x,
        )

    def on_inference_end(
        self,
        y_pred: Tensor,
        y_true: Tensor,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_inference_end' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param y_pred:    The prediction (output) of the model for this batch's input ('x').
        :param y_true:    The true value part of the current batch.
        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_INFERENCE_END,
            callbacks=self._parse_names(names=callbacks),
            y_pred=y_pred,
            y_true=y_true,
        )

    def on_train_loss_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_train_loss_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_LOSS_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_train_loss_end(
        self,
        loss_value: PyTorchTypes.MetricValueType,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_train_loss_end' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param loss_value: The recent loss value calculated during training.
        :param callbacks:  The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_LOSS_END,
            callbacks=self._parse_names(names=callbacks),
            loss_value=loss_value,
        )

    def on_validation_loss_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_validation_loss_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_LOSS_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_validation_loss_end(
        self,
        loss_value: PyTorchTypes.MetricValueType,
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_validation_loss_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param loss_value: The recent loss value calculated during validation.
        :param callbacks:  The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_LOSS_END,
            callbacks=self._parse_names(names=callbacks),
            loss_value=loss_value,
        )

    def on_train_metrics_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_train_metrics_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_METRICS_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_train_metrics_end(
        self,
        metric_values: list[PyTorchTypes.MetricValueType],
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_train_metrics_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param metric_values: The recent metric values calculated during training.
        :param callbacks:     The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_TRAIN_METRICS_END,
            callbacks=self._parse_names(names=callbacks),
            metric_values=metric_values,
        )

    def on_validation_metrics_begin(
        self, callbacks: Optional[list[str]] = None
    ) -> bool:
        """
        Call the 'on_validation_metrics_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_METRICS_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_validation_metrics_end(
        self,
        metric_values: list[PyTorchTypes.MetricValueType],
        callbacks: Optional[list[str]] = None,
    ) -> bool:
        """
        Call the 'on_validation_metrics_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param metric_values: The recent metric values calculated during validation.
        :param callbacks:     The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_VALIDATION_METRICS_END,
            callbacks=self._parse_names(names=callbacks),
            metric_values=metric_values,
        )

    def on_backward_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_backward_begin' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_BACKWARD_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_backward_end(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_backward_end' method of every callback in the callbacks list. If the list is 'None' (not given),
        all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_BACKWARD_END,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_optimizer_step_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_optimizer_step_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_OPTIMIZER_STEP_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_optimizer_step_end(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_optimizer_step_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_OPTIMIZER_STEP_END,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_scheduler_step_begin(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_scheduler_step_begin' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_SCHEDULER_STEP_BEGIN,
            callbacks=self._parse_names(names=callbacks),
        )

    def on_scheduler_step_end(self, callbacks: Optional[list[str]] = None) -> bool:
        """
        Call the 'on_scheduler_step_end' method of every callback in the callbacks list. If the list is 'None'
        (not given), all callbacks will be called.

        :param callbacks: The callbacks names to use. If 'None', all of the callbacks will be used.

        :return: True if all of the callbacks called returned True and False if not.
        """
        return self._run_callbacks(
            method_name=_CallbackInterface.ON_SCHEDULER_STEP_END,
            callbacks=self._parse_names(names=callbacks),
        )

    def _parse_names(self, names: Union[list[str], None]) -> list[str]:
        """
        Parse the given callbacks names. If they are not 'None' then the names will be returned as they are, otherwise
        all of the callbacks handled by this handler will be returned (the default behavior of when there were no names
        given to one of the handler's methods).

        :param names: A list of names to parse, can be 'None'.

        :return: The given names if they were not 'None', otherwise the names of all the callbacks in this handler.
        """
        if names:
            return names
        return list(self._callbacks.keys())

    def _run_callbacks(
        self, method_name: str, callbacks: list[str], *args, **kwargs
    ) -> bool:
        """
        Run the given method from the 'CallbackInterface' on all the specified callbacks with the given arguments.

        :param method_name: The name of the method to run. Should be given from the 'CallbackInterface'.
        :param callbacks:   List of all the callbacks names to run the method.

        :return: True if all the callbacks called returned True and False if not.
        """
        all_result = True
        for callback in callbacks:
            if self._callbacks[callback].on_call_check():
                method = getattr(self._callbacks[callback], method_name)
                result = method(*args, **kwargs)
                if result:
                    all_result &= result
        return all_result



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import MLTypes


class Logger:
    """
    Logger for tracking hyperparamters and metrics results during training of some framework.
    """

    def __init__(self):
        """
        Initialize a generic logger for collecting training data.
        """
        # Set up the results dictionaries - a dictionary of metrics for all the iteration results by their epochs:
        # [Validation Set: str] -> [Metric: str] -> [Iteration: int] -> [value: float]
        self._results = {}  # type: Dict[str, Dict[str, List[float]]]

        # Store the static hyperparameters given - a dictionary of parameters and their values to note:
        # [Parameter: str] -> [value: Union[str, bool, float, int]]
        self._static_hyperparameters = {}  # type: Dict[str, MLTypes.TrackableType]

        # Set up the dynamic hyperparameters dictionary - a dictionary of all tracked hyperparameters by epochs:
        # [Hyperparameter: str] -> [Epoch: int] -> [value: Union[str, bool, float, int]]
        self._dynamic_hyperparameters = {}  # type: Dict[str, List[MLTypes.TrackableType]]

        # Set up the iterations counter:
        self._iterations = 0

    @property
    def results(self) -> dict[str, dict[str, list[float]]]:
        """
        Get the results logged. The results will be stored in a dictionary where each key is the validation set name
        and the value is a dictionary of metrics to their list of iterations values.

        :return: The results.
        """
        return self._results

    @property
    def static_hyperparameters(self) -> dict[str, MLTypes.TrackableType]:
        """
        Get the static hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is his logged value.

        :return: The static hyperparameters.
        """
        return self._static_hyperparameters

    @property
    def dynamic_hyperparameters(self) -> dict[str, list[MLTypes.TrackableType]]:
        """
        Get the dynamic hyperparameters logged. The hyperparameters will be stored in a dictionary where each key is the
        hyperparameter name and the value is a list of his logged values per epoch.

        :return: The dynamic hyperparameters.
        """
        return self._dynamic_hyperparameters

    @property
    def iterations(self) -> int:
        """
        Get the overall iterations.

        :return: The overall iterations.
        """
        return self._iterations

    def log_iteration(self):
        """
        Log a new iteration.
        """
        self._iterations += 1

    def log_result(self, validation_set_name: str, metric_name: str, result: float):
        """
        Log the given metric result in the training results dictionary at the current epoch.

        :param validation_set_name: Name of the validation set used.
        :param metric_name:         The metric name.
        :param result:              The metric result to log.
        """
        # Get the validation set's metrics (will set a new dictionary in case it's a new validation set):
        if validation_set_name not in self._results:
            self._results[validation_set_name] = {}
        validation_set_metrics = self._results[validation_set_name]

        # Get the metric's results list (will set a new list in case it's a new metric):
        if metric_name not in validation_set_metrics:
            validation_set_metrics[metric_name] = []
        metric_results = validation_set_metrics.setdefault(metric_name, [])

        # Log the metric's result:
        metric_results.append(result)

    def log_static_hyperparameter(
        self, parameter_name: str, value: MLTypes.TrackableType
    ):
        """
        Log the given parameter value in the static hyperparameters dictionary.

        :param parameter_name: The parameter name.
        :param value:          The parameter value to log.
        """
        self._static_hyperparameters[parameter_name] = value

    def log_dynamic_hyperparameter(
        self, parameter_name: str, value: MLTypes.TrackableType
    ):
        """
        Log the given parameter value in the dynamic hyperparameters dictionary at the current iteration (if it's a new
        parameter it will be iteration 0). If the parameter appears in the static hyperparameters dictionary, it will be
        removed from there as it is now dynamic.

        :param parameter_name: The parameter name.
        :param value:          The parameter value to log.
        """
        # Check if it's a new hyperparameter being tracked:
        if parameter_name not in self._dynamic_hyperparameters:
            # Look in the static hyperparameters:
            if parameter_name in self._static_hyperparameters:
                self._static_hyperparameters.pop(parameter_name)
            # Add it as a dynamic hyperparameter:
            self._dynamic_hyperparameters[parameter_name] = [value]
        else:
            self._dynamic_hyperparameters[parameter_name].append(value)



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Awaitable
from typing import Any, Callable, Optional, Union

import mlrun.errors
from mlrun.common.types import StrEnum
from mlrun.datastore.remote_client import (
    BaseRemoteClient,
)


class InvokeResponseFormat(StrEnum):
    STRING = "string"
    USAGE = "usage"
    FULL = "full"

    @classmethod
    def is_str_response(cls, invoke_response_format: str) -> bool:
        """
        Returns True if the response key corresponds to a string-based response (not a full generation object).
        """
        return invoke_response_format in {
            cls.USAGE,
            cls.STRING,
        }


class UsageResponseKeys(StrEnum):
    ANSWER = "answer"
    USAGE = "usage"

    @classmethod
    def fields(cls) -> list[str]:
        return [cls.ANSWER, cls.USAGE]


class ModelProvider(BaseRemoteClient):
    """
    The ModelProvider class is an abstract base for integrating with external
    model providers, primarily generative AI (GenAI) services.

    Designed to be subclassed, it defines a consistent interface and shared
    functionality for tasks such as text generation, embeddings, and invoking
    fine-tuned models. Subclasses should implement provider-specific logic,
    including SDK client initialization, model invocation, and custom operations.

    Key Features:
    - Establishes a consistent, reusable client management for model provider integrations.
    - Simplifies GenAI service integration by abstracting common operations.
    - Reduces duplication through shared components for common tasks.
    - Holds default invocation parameters (e.g., temperature, max_tokens) to avoid boilerplate
    code and promote consistency.
    """

    support_async = False

    def __init__(
        self,
        parent,
        kind,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        super().__init__(
            parent=parent, name=name, kind=kind, endpoint=endpoint, secrets=secrets
        )
        self.default_invoke_kwargs = default_invoke_kwargs or {}
        self._client = None
        self._async_client = None

    @staticmethod
    def _extract_string_output(response: Any) -> str:
        """
        Extracts string response from response object
        """
        pass

    def _response_handler(
        self,
        response: Any,
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **kwargs,
    ) -> Union[str, dict, Any]:
        """
        Handles the model response according to the specified response format.

        :param response: The raw response returned from the model invocation.
        :param invoke_response_format: Determines how the response should be processed and returned.
                                       Options include:

                                       - STRING: Return only the main generated content as a string,
                                                 typically for single-answer responses.
                                       - USAGE: Return a dictionary combining the string response with
                                                additional metadata or token usage statistics, in this format:
                                                {"answer": <string>, "usage": <dict>}

                                       - FULL: Return the full raw response object.

        :param kwargs:                  Additional parameters that may be required by specific implementations.

        :return:                        The processed response in the format specified by `invoke_response_format`.
                                        Can be a string, dictionary, or the original response object.
        """
        return None

    def get_client_options(self) -> dict:
        """
        Returns a dictionary containing credentials and configuration
        options required for client creation.

        :return:           A dictionary with client-specific settings.
        """
        return {}

    def load_client(self) -> None:
        """
        Initialize the SDK client for the model provider and assign it to an instance attribute.

        Subclasses should override this method to create and configure the provider-specific client.
        """

        raise NotImplementedError("load_client method is not implemented")

    def load_async_client(self) -> Any:
        raise NotImplementedError("load_async_client method is not implemented")

    @property
    def client(self) -> Any:
        return self._client

    @property
    def model(self) -> Optional[str]:
        """
        Returns the model identifier used by the underlying SDK.

        :return: A string representing the model ID, or None if not set.
        """
        return self.endpoint

    def get_invoke_kwargs(self, invoke_kwargs) -> dict:
        kwargs = self.default_invoke_kwargs.copy()
        kwargs.update(invoke_kwargs)
        return kwargs

    @property
    def async_client(self) -> Any:
        if not self.support_async:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"{self.__class__.__name__} does not support async operations"
            )
        return self._async_client

    def custom_invoke(
        self, operation: Optional[Callable] = None, **invoke_kwargs
    ) -> Any:
        """
        Invokes a model operation from a provider (e.g., OpenAI, Hugging Face, etc.) with the given keyword arguments.

        Useful for dynamically calling model methods like text generation, chat completions, or image generation.
        The operation must be a callable that accepts keyword arguments.

        :param operation:       A callable representing the model operation (e.g., a client method).
        :param invoke_kwargs:   Keyword arguments to pass to the operation.
        :return:                The full response returned by the operation.
        """
        raise NotImplementedError("custom_invoke method is not implemented")

    async def async_custom_invoke(
        self, operation: Optional[Callable[..., Awaitable[Any]]] = None, **invoke_kwargs
    ) -> Any:
        """
        Asynchronously invokes a model operation from a provider (e.g., OpenAI, Hugging Face, etc.)
        with the given keyword arguments.

        The operation must be an async callable (e.g., a method from an async client) that accepts keyword arguments.

        :param operation:       An async callable representing the model operation (e.g., an async_client method).
        :param invoke_kwargs:   Keyword arguments to pass to the operation.
        :return:                The full response returned by the awaited operation.
        """
        raise NotImplementedError("async_custom_invoke is not implemented")

    def invoke(
        self,
        messages: Union[list[dict], Any],
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[str, dict[str, Any], Any]:
        """
        Invokes a generative AI model with the provided messages and additional parameters.
        This method is designed to be a flexible interface for interacting with various
        generative AI backends (e.g., OpenAI, Hugging Face, etc.). It allows users to send
        a list of messages (following a standardized format) and receive a response.

        :param messages:            A list of dictionaries representing the conversation history or input messages.
                                    Each dictionary should follow the format::
                                    {"role": "system"| "user" | "assistant" ..., "content":
                                    "Message content as a string"}

                                    Example:

                                    .. code-block:: json

                                        [
                                            {"role": "system", "content": "You are a helpful assistant."},
                                            {"role": "user", "content": "What is the capital of France?"}
                                        ]

                                    This format is consistent across all backends. Defaults to None if no messages
                                    are provided.

        :param invoke_response_format:   Determines how the model response is returned:

                                    - string:   Returns only the generated text content from the model output,
                                                for single-answer responses only.

                                    - usage:    Combines the STRING response with additional metadata (token usage),
                                                and returns the result in a dictionary.

                                                Note: The usage dictionary may contain additional
                                                keys depending on the model provider:

                                    .. code-block:: json

                                    {
                                        "answer": "<generated_text>",
                                        "usage": {
                                        "prompt_tokens": <int>,
                                        "completion_tokens": <int>,
                                        "total_tokens": <int>
                                        }

                                    }

                                    - full:   Returns the full model output.

        :param invoke_kwargs:
                                    Additional keyword arguments to be passed to the underlying model API call.
                                    These can include parameters such as temperature, max tokens, etc.,
                                    depending on the capabilities of the specific backend being used.

        :return:                    The invoke result formatted according to the specified
                                    invoke_response_format parameter.

        """
        raise NotImplementedError("invoke method is not implemented")

    async def async_invoke(
        self,
        messages: list[dict],
        invoke_response_format=InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[str, dict[str, Any], Any]:
        """
        Asynchronously invokes a generative AI model with the provided messages and additional parameters.
        This method is designed to be a flexible interface for interacting with various
        generative AI backends (e.g., OpenAI, Hugging Face, etc.). It allows users to send
        a list of messages (following a standardized format) and receive a response.

        :param messages:            A list of dictionaries representing the conversation history or input messages.
                                    Each dictionary should follow the format::
                                    {"role": "system"| "user" | "assistant" ..., "content":
                                    "Message content as a string"}

                                    Example:

                                    .. code-block:: json

                                        [
                                            {"role": "system", "content": "You are a helpful assistant."},
                                            {"role": "user", "content": "What is the capital of France?"}
                                        ]

                                    This format is consistent across all backends. Defaults to None if no messages
                                    are provided.

        :param invoke_response_format:   Determines how the model response is returned:

                                    - string:   Returns only the generated text content from the model output,
                                                for single-answer responses only.

                                    - usage:    Combines the STRING response with additional metadata (token usage),
                                                and returns the result in a dictionary.

                                                Note: The usage dictionary may contain additional
                                                keys depending on the model provider:

                                    .. code-block:: json

                                    {
                                        "answer": "<generated_text>",
                                        "usage": {
                                        "prompt_tokens": <int>,
                                        "completion_tokens": <int>,
                                        "total_tokens": <int>
                                        }

                                    }

                                    - full:   Returns the full model output.

        :param invoke_kwargs:
                                    Additional keyword arguments to be passed to the underlying model API call.
                                    These can include parameters such as temperature, max tokens, etc.,
                                    depending on the capabilities of the specific backend being used.

        :return:                    The invoke result formatted according to the specified
                                    invoke_response_format parameter.

        """
        raise NotImplementedError("async_invoke is not implemented")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pipeline param format which is passed when running a pipeline (e.g. {{pipelineparam:op=;name=mem}})
# https://github.com/kubeflow/pipelines/blob/16edebf4eaf84cd7478e2601ef4878ab339a7854/sdk/python/kfp/dsl/_pipeline_param.py#L213
# this is expected to be resolved at runtime
pipeline_param = [r"{{pipelineparam:op=([\w\s_-]*);name=([\w\s_-]+)}}"]

# k8s character limit is for 63 characters
k8s_character_limit = [r"^.{0,63}$"]

# k8s name
# https://github.com/kubernetes/apimachinery/blob/kubernetes-1.25.16/pkg/util/validation/validation.go#L33
qualified_name = [r"^(([A-Za-z0-9][-A-Za-z0-9_.]*)?[A-Za-z0-9])?$"]

# k8s label value format
# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/util/validation/validation.go#L161
label_value = k8s_character_limit + qualified_name

# DNS Subdomain (RFC 1123) - used by k8s for most resource names format
# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/util/validation/validation.go#L204
dns_1123_subdomain = [
    r"^.{0,253}$",
    r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*$",
]

# DNS Label (RFC 1123) - used by k8s for resource names format
# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/util/validation/validation.go#L183
dns_1123_label = k8s_character_limit + [r"^[a-z0-9]([-a-z0-9]*[a-z0-9])?$"]

# DNS 1035 - used by k8s for services services
# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/util/validation/validation.go#L220
dns_1035_label = [r"[a-z]([-a-z0-9]*[a-z0-9])?"]

# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/util/validation/validation.go#L424
k8s_secret_and_config_map_key = [
    r"^.{0,253}$",
    r"^[-._a-zA-Z0-9]+$",
]

# https://github.com/kubernetes/kubernetes/blob/v1.20.0/staging/src/k8s.io/apimachinery/pkg/api/resource/quantity.go#L136
k8s_resource_quantity_regex = [r"^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$"]

run_name = label_value

# Sparkjob name value format
# The actual limit is 63 characters, but due to mlrun and spark operator additions for unique
# values - the limit is set to 30.
# The names of the generated resources are in the format: [function_name]-[uid*8]-[generated_resource_name]
#   function_name - the name provided by the user
#   uid*8 - is 8 characters generated in mlrun to give the resources a guaranteed unique name.
#   generated_resource_name - for each resource added a describing suffix (e.g. driver) in that case
#       the longest suffix is "-[uid*13]-driver-svc" which contains 25 characters. Therefore the limit should be
#       63 - 25 - 9 = 29
#       NOTE: If a name is between 30-33 characters - the function will complete successfully without creating the
#           driver-svc meaning there is no way to get the response through a ui
sprakjob_length = [r"^.{0,29}$"]

# part of creating sparkjob operator it creates a service with the same name of the job
sparkjob_service_name = dns_1035_label

sparkjob_name = label_value + sprakjob_length + sparkjob_service_name

# A project name have the following restrictions:
# It should be a valid Nuclio Project CRD name which is dns 1123 subdomain
# It should be a valid k8s label value since Nuclio use the project name in labels of resources
# It should be a valid namespace name (because we plan to map it to one) which is dns 1123 label
# of the 3 restrictions, dns 1123 label is strictest, so we enforce only it
project_name = dns_1123_label

# Special characters are not permitted in tag names because they can be included in the url and cause problems.
# We only accept letters, capital letters, numbers, dots, and hyphens, with a k8s character limit.
tag_name = label_value

secret_key = k8s_secret_and_config_map_key

artifact_key = [r"^[A-Za-z0-9]([-A-Za-z0-9_.]*[A-Za-z0-9])?$"]

# must not start with _
# must be alphanumeric or _
# max 256 length
v3io_stream_consumer_group = [r"^(?!_)[a-zA-Z0-9_]{1,256}$"]

# URI patterns
run_uri_pattern = r"^(?P<project>.*)@(?P<uid>.*)\#(?P<iteration>.*?)(:(?P<tag>.*))?$"

artifact_uri_pattern = (
    r"^((?P<project>.*)/)?"  # Optional project
    r"(?P<key>.*?)"  # Key
    r"(\#(?P<iteration>.*?))?"  # Optional iteration
    r"(:(?P<tag>.*?))?"  # Optional tag
    r"(@(?P<tree>.*?))?"  # Optional tree
    r"(\^(?P<uid>.*))?$"  # Optional uid
)

artifact_producer_uri_pattern = (
    r"^((?P<project>.*)/)?(?P<uid>.*?)(\-(?P<iteration>.*?))?$"
)

mail_regex = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"
alert_name_regex = r"^[a-zA-Z0-9-_]+$"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
import pandas as pd
from cloudpickle import load

from mlrun.serving.v2_serving import V2ModelServer


class PickleModelServer(V2ModelServer):
    """
    Model serving class, inheriting the V2ModelServer class for being initialized automatically by the model
    server and be able to run locally as part of a nuclio serverless function, or as part of a real-time pipeline.
    """

    def load(self):
        """
        Load and initialize the model and/or other elements.
        """
        model_file, extra_data = self.get_model(".pkl")
        self.model = load(open(model_file, "rb"))

    def predict(self, request: dict) -> list:
        """
        Infer the inputs through the model using MLRun's PyTorch interface and return its output. The inferred data will
        be read from the "inputs" key of the request.

        :param request: The request of the model. The input to the model will be read from the "inputs" key.
                        The input value can be either a dictionary with the feature names
                        (only one can be sent so only the first one will take into consideration)
                        or a list with feature values.
                        For example, a batch size of 2 for data of two features 'x' and 'y' can be given:

                        * As a dictionary: `{"inputs": [{"x": [1, 2], "y": [3, 5.5]}]}`
                        * As a list: `{"inputs": [[1, 2], [3, 5.5]]}`
        :return: The model's prediction on the given input.
        """
        inputs = request["inputs"]
        if inputs and isinstance(inputs[0], dict):
            x = pd.DataFrame(inputs[0])
        else:
            x = np.asarray(inputs)

        y_pred: np.ndarray = self.model.predict(x)

        return y_pred.tolist()

    def explain(self, request: dict[str, Any]) -> str:
        """
        Returns a string listing the model that is being served in this serving function and the function name.

        :param request: A given request.

        :return: Explanation string.
        """
        return f"A model server named '{self.name}'"



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from datetime import datetime
from types import TracebackType
from typing import TYPE_CHECKING, Final, Optional

import botocore.exceptions

import mlrun
import mlrun.common.schemas as schemas
import mlrun.errors
import mlrun.model_monitoring.helpers
import mlrun.utils.helpers
from mlrun.utils import logger

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class ModelMonitoringSchedulesFileBase(AbstractContextManager, ABC):
    DEFAULT_SCHEDULES: Final = {}
    INITIAL_CONTENT = json.dumps(DEFAULT_SCHEDULES)
    ENCODING = "utf-8"

    def __init__(self) -> None:
        # `self._item` is the persistent version of the monitoring schedules.
        self._item = self.get_data_item_object()
        if self._item:
            self._path = self._item.url
            self._fs = self._item.store.filesystem
            # `self._schedules` is an in-memory copy of the DB for all the applications for
            # the same model endpoint.
            self._schedules = self.DEFAULT_SCHEDULES.copy()
            # Does `self._schedules` hold the content of `self._item`?
            self._open_schedules = False

    @abstractmethod
    def get_data_item_object(self) -> "mlrun.DataItem":
        pass

    def _exists(self) -> bool:
        """Return whether the file exists or not"""
        return (
            self._fs is None  # In-memory store
            or self._fs.exists(self._path)
        )

    def create(self) -> None:
        """Create a schedules file with initial content - an empty dictionary"""
        logger.debug("Creating model monitoring schedules file", path=self._item.url)
        self._item.put(self.INITIAL_CONTENT)

    def delete(self) -> None:
        """Delete schedules file if it exists"""
        if self._exists():
            logger.debug(
                "Deleting model monitoring schedules file", path=self._item.url
            )
            self._item.delete()
        else:
            logger.debug(
                "Model monitoring schedules file does not exist, nothing to delete",
                path=self._item.url,
            )

    def _open(self) -> None:
        try:
            content = self._item.get()
        except (
            mlrun.errors.MLRunNotFoundError,
            # Different errors are raised for S3 or local storage, see ML-8042
            botocore.exceptions.ClientError,
            FileNotFoundError,
        ) as err:
            if (
                isinstance(err, botocore.exceptions.ClientError)
                # Add a log only to "NoSuchKey" errors codes - equivalent to `FileNotFoundError`
                and err.response["Error"]["Code"] != "NoSuchKey"
            ):
                raise

            logger.exception(
                "The schedules file was not found. It should have been created "
                "as a part of the model endpoint's creation",
                path=self._path,
            )
            raise

        if isinstance(content, bytes):
            content = content.decode(encoding=self.ENCODING)
        self._schedules = json.loads(content)
        self._open_schedules = True

    def _close(self) -> None:
        self._item.put(json.dumps(self._schedules))
        self._schedules = self.DEFAULT_SCHEDULES
        self._open_schedules = False

    def __enter__(self) -> "Self":
        self._open()
        return super().__enter__()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self._close()

    def _check_open_schedules(self) -> None:
        if not self._open_schedules:
            raise mlrun.errors.MLRunValueError(
                "Open the schedules file as a context manager first"
            )


class ModelMonitoringSchedulesFileEndpoint(ModelMonitoringSchedulesFileBase):
    def __init__(self, project: str, endpoint_id: str) -> None:
        """
        Initialize applications monitoring schedules file object.
        The JSON file stores a dictionary of registered application name as key and Unix timestamp as value.
        When working with the schedules data, use this class as a context manager to read and write the data.

        :param project:     The project name.
        :param endpoint_id: The endpoint ID.
        """
        self._project = project
        self._endpoint_id = endpoint_id
        super().__init__()

    def get_data_item_object(self) -> "mlrun.DataItem":
        return mlrun.model_monitoring.helpers.get_monitoring_schedules_endpoint_data(
            project=self._project, endpoint_id=self._endpoint_id
        )

    @classmethod
    def from_model_endpoint(
        cls, model_endpoint: schemas.ModelEndpoint
    ) -> "ModelMonitoringSchedulesFileEndpoint":
        return cls(
            project=model_endpoint.metadata.project,
            endpoint_id=model_endpoint.metadata.uid,
        )

    def get_application_time(self, application: str) -> Optional[float]:
        self._check_open_schedules()
        return self._schedules.get(application)

    def update_application_time(self, application: str, timestamp: float) -> None:
        self._check_open_schedules()
        self._schedules[application] = float(timestamp)

    def delete_application_time(self, application: str) -> None:
        self._check_open_schedules()
        if application in self._schedules:
            logger.debug(
                "Deleting application time from schedules",
                application=application,
                endpoint_id=self._endpoint_id,
            )
            del self._schedules[application]

    def get_application_list(self) -> set[str]:
        self._check_open_schedules()
        return set(self._schedules.keys())

    def get_min_timestamp(self) -> Optional[float]:
        self._check_open_schedules()
        return min(self._schedules.values(), default=None)


class ModelMonitoringSchedulesFileChief(ModelMonitoringSchedulesFileBase):
    def __init__(self, project: str) -> None:
        """
        Initialize applications monitoring schedules chief file object.
        The JSON file stores a dictionary of registered model endpoints uid as key and point to a dictionary of
        "last_request" and "last_analyzed" mapped to two Unix timestamps as values.
        When working with the schedules data, use this class as a context manager to read and write the data.

        :param project:     The project name.
        """
        # `self._item` is the persistent version of the monitoring schedules.
        self._project = project
        super().__init__()

    def get_data_item_object(self) -> "mlrun.DataItem":
        return mlrun.model_monitoring.helpers.get_monitoring_schedules_chief_data(
            project=self._project
        )

    def get_endpoint_last_request(self, endpoint_uid: str) -> Optional[float]:
        self._check_open_schedules()
        if endpoint_uid in self._schedules:
            return self._schedules[endpoint_uid].get(
                schemas.model_monitoring.constants.ScheduleChiefFields.LAST_REQUEST
            )
        else:
            return None

    def update_endpoint_timestamps(
        self, endpoint_uid: str, last_request: float, last_analyzed: float
    ) -> None:
        self._check_open_schedules()
        self._schedules[endpoint_uid] = {
            schemas.model_monitoring.constants.ScheduleChiefFields.LAST_REQUEST: float(
                last_request
            ),
            schemas.model_monitoring.constants.ScheduleChiefFields.LAST_ANALYZED: float(
                last_analyzed
            ),
        }

    def get_endpoint_last_analyzed(self, endpoint_uid: str) -> Optional[float]:
        self._check_open_schedules()
        if endpoint_uid in self._schedules:
            return self._schedules[endpoint_uid].get(
                schemas.model_monitoring.constants.ScheduleChiefFields.LAST_ANALYZED
            )
        else:
            return None

    def get_endpoint_list(self) -> set[str]:
        self._check_open_schedules()
        return set(self._schedules.keys())

    def get_or_create(self) -> None:
        if not self._exists():
            self.create()


class ModelMonitoringSchedulesFileApplication(ModelMonitoringSchedulesFileBase):
    def __init__(self, out_path: str, application: str) -> None:
        self._out_path = out_path
        self._application = application
        super().__init__()

    def get_data_item_object(self) -> "mlrun.DataItem":
        return mlrun.model_monitoring.helpers.get_monitoring_schedules_user_application_data(
            out_path=self._out_path, application=self._application
        )

    def _open(self) -> None:
        if not self._exists():
            # Create the file when it is needed the first time
            logger.info(
                "Creating the application schedules file",
                application=self._application,
                path=self._path,
            )
            self.create()
        super()._open()

    def get_endpoint_last_analyzed(self, endpoint_uid: str) -> Optional[datetime]:
        self._check_open_schedules()
        if endpoint_uid in self._schedules:
            return datetime.fromisoformat(self._schedules[endpoint_uid])
        else:
            return None

    def update_endpoint_last_analyzed(
        self, endpoint_uid: str, last_analyzed: datetime
    ) -> None:
        self._check_open_schedules()
        self._schedules[endpoint_uid] = last_analyzed.isoformat()

    def delete_endpoints_last_analyzed(self, endpoint_uids: list[str]) -> None:
        self._check_open_schedules()
        for endpoint_uid in endpoint_uids:
            if endpoint_uid in self._schedules:
                logger.debug(
                    "Deleting endpoint last analyzed from schedules",
                    endpoint_uid=endpoint_uid,
                    application=self._application,
                )
                del self._schedules[endpoint_uid]


def _delete_folder(folder: str) -> None:
    fs = mlrun.datastore.store_manager.object(folder).store.filesystem
    if fs and fs.exists(folder):
        logger.debug("Deleting model monitoring schedules folder", folder=folder)
        fs.rm(folder, recursive=True)
    elif fs is None:  # In-memory store
        raise mlrun.errors.MLRunValueError(
            "Cannot delete a folder without a file-system"
        )


def delete_model_monitoring_schedules_folder(project: str) -> None:
    """Delete the model monitoring schedules folder of the project"""
    folder = mlrun.model_monitoring.helpers._get_monitoring_schedules_folder_path(
        project
    )
    _delete_folder(folder)


def delete_model_monitoring_schedules_user_folder(project: str) -> None:
    """Delete the user created schedules folder (created through `app.evaluate`)"""
    out_path = mlrun.utils.helpers.template_artifact_path(
        mlrun.mlconf.artifact_path, project=project
    )
    folder = mlrun.model_monitoring.helpers._get_monitoring_schedules_user_folder_path(
        out_path
    )
    _delete_folder(folder)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
from collections import OrderedDict
from typing import Union

from mlrun.datastore import DataItem
from mlrun.errors import MLRunInvalidArgumentError
from mlrun.execution import MLClientCtx
from mlrun.run import get_or_create_ctx

from .errors import MLRunPackageCollectionError, MLRunPackagePackingError
from .packagers_manager import PackagersManager
from .utils import LogHintUtils, TypeHintUtils


class ContextHandler:
    """
    A class for handling a MLRun context of a function that is wrapped in MLRun's `handler` decorator.

    The context handler have 3 duties:
      1. Check if the user used MLRun to run the wrapped function and if so, get the MLRun context.
      2. Parse the user's inputs (MLRun `DataItem`) to the function.
      3. Log the function's outputs to MLRun.

    The context handler uses a packagers manager to unpack (parse) the inputs and pack (log) the outputs. It sets up a
    manager with all the packagers in the `mlrun.package.packagers` directory. Packagers whom are in charge of modules
    that are in the MLRun requirements are mandatory and additional extensions packagers for non-required modules are
    added if the modules are available in the user's interpreter. Once a context is found, project custom packagers will
    be added as well.
    """

    # Mandatory packagers to be collected at initialization time:
    _MLRUN_REQUIREMENTS_PACKAGERS = [
        "python_standard_library",
        "pandas",
        "numpy",
    ]
    # Optional packagers to be collected at initialization time:
    _EXTENDED_PACKAGERS = []  # TODO: Create "matplotlib", "plotly", packagers.
    # Optional packagers from the `mlrun.frameworks` package:
    _MLRUN_FRAMEWORKS_PACKAGERS = []  # TODO: Create frameworks packagers.
    # Default priority values for packagers:
    _BUILTIN_PACKAGERS_DEFAULT_PRIORITY = 5
    _CUSTOM_PACKAGERS_DEFAULT_PRIORITY = 3

    def __init__(self):
        """
        Initialize a context handler.
        """
        # Set up a variable to hold the context:
        self._context: MLClientCtx = None

        # Initialize a packagers manager:
        self._packagers_manager = PackagersManager()

        # Prepare the manager (collect the MLRun builtin standard and optional packagers):
        self._collect_mlrun_packagers()

    def look_for_context(self, args: tuple, kwargs: dict):
        """
        Look for an MLRun context (`mlrun.MLClientCtx`). The handler will look for a context in the given order:
          1. The given arguments.
          2. The given keyword arguments.
          3. If an MLRun RunTime was used the context will be located via the `mlrun.get_or_create_ctx` method.

        :param args:   The arguments tuple passed to the function.
        :param kwargs: The keyword arguments dictionary passed to the function.
        """
        # Search in the given arguments:
        for argument in args:
            if isinstance(argument, MLClientCtx):
                self._context = argument
                break

        # Search in the given keyword arguments:
        if self._context is None:
            for argument_name, argument_value in kwargs.items():
                if isinstance(argument_value, MLClientCtx):
                    self._context = argument_value
                    break

        # Search if the function was triggered from an MLRun RunTime object by looking at the call stack:
        # Index 0: the current frame.
        # Index 1: the decorator's frame.
        # Index 2-...: If it is from mlrun.runtimes we can be sure it ran via MLRun, otherwise not.
        if self._context is None:
            for callstack_frame in inspect.getouterframes(inspect.currentframe()):
                if (
                    os.path.join("mlrun", "runtimes", "local")
                    in callstack_frame.filename
                ):
                    self._context = get_or_create_ctx("context")
                    break

        # Give the packagers manager custom packagers to collect (if a context is found and a project is available):
        if self._context is not None and self._context.project:
            # Get the custom packagers property from the project's spec:
            project = self._context.get_project_object()
            if project and project.spec.custom_packagers:
                # Add the custom packagers taking into account the mandatory flag:
                for custom_packager, is_mandatory in project.spec.custom_packagers:
                    self._collect_packagers(
                        packagers=[custom_packager],
                        is_mandatory=is_mandatory,
                        is_custom_packagers=True,
                    )

    def is_context_available(self) -> bool:
        """
        Check if a context was found by the method `look_for_context`.

        :returns: True if a context was found and False otherwise.
        """
        return self._context is not None

    def parse_inputs(
        self,
        args: tuple,
        kwargs: dict,
        type_hints: OrderedDict,
    ) -> tuple:
        """
        Parse the given arguments and keyword arguments data items to the expected types.

        :param args:       The arguments tuple passed to the function.
        :param kwargs:     The keyword arguments dictionary passed to the function.
        :param type_hints: An ordered dictionary of the expected types of arguments.

        :returns: The parsed args (kwargs are parsed inplace).
        """
        # Parse the type hints (in case some were given as strings):
        type_hints = {
            key: TypeHintUtils.parse_type_hint(type_hint=value)
            for key, value in type_hints.items()
        }

        # Parse the arguments:
        parsed_args = []
        type_hints_keys = list(type_hints.keys())
        for i, argument in enumerate(args):
            if (
                isinstance(argument, DataItem)
                and type_hints[type_hints_keys[i]] is not inspect.Parameter.empty
            ):
                parsed_args.append(
                    self._packagers_manager.unpack(
                        data_item=argument,
                        type_hint=type_hints[type_hints_keys[i]],
                    )
                )
            else:
                parsed_args.append(argument)
        parsed_args = tuple(parsed_args)  # `args` is expected to be a tuple.

        # Parse the keyword arguments:
        for key, value in kwargs.items():
            if (
                isinstance(value, DataItem)
                and type_hints[key] is not inspect.Parameter.empty
            ):
                kwargs[key] = self._packagers_manager.unpack(
                    data_item=value, type_hint=type_hints[key]
                )

        return parsed_args

    def log_outputs(
        self,
        outputs: list,
        log_hints: list[Union[dict[str, str], str, None]],
    ):
        """
        Log the given outputs as artifacts (or results) with the stored context. Errors raised during the packing will
        be ignored to not fail a run. A warning with the error wil be printed.

        Only the logging worker will pack and log the outputs.

        :param outputs:   List of outputs to log.
        :param log_hints: List of log hints (logging configurations) to use.
        """
        # Pack and log only from the logging worker (in case of multi-workers job like OpenMPI):
        if self._context.is_logging_worker():
            # Verify the outputs and log hints are the same length:
            self._validate_objects_to_log_hints_length(
                outputs=outputs, log_hints=log_hints
            )
            # Go over the outputs and pack them:
            for obj, log_hint in zip(outputs, log_hints):
                try:
                    # Check if needed to log (not None):
                    if log_hint is None:
                        continue
                    # Parse the log hint:
                    log_hint = LogHintUtils.parse_log_hint(log_hint=log_hint)
                    # Pack the object (we don't catch the returned package as we log it after we pack all the outputs to
                    # enable linking extra data of some artifacts):
                    self._packagers_manager.pack(obj=obj, log_hint=log_hint)
                except (MLRunInvalidArgumentError, MLRunPackagePackingError) as error:
                    self._context.logger.warn(
                        f"Skipping logging an object with the log hint '{log_hint}' "
                        f"due to the following error:\n{error}"
                    )
            # Link packages:
            self._packagers_manager.link_packages(
                additional_artifact_uris=self._context.artifact_uris,
                additional_results=self._context.results,
            )
            # Log the packed results and artifacts:
            self._context.log_results(results=self._packagers_manager.results)
            for artifact in self._packagers_manager.artifacts:
                self._context.log_artifact(item=artifact)
        else:
            self._context.logger.debug("Skipping logging - not the logging worker.")

        # Clear packagers outputs:
        self._packagers_manager.clear_packagers_outputs()

    def set_labels(self, labels: dict[str, str]):
        """
        Set the given labels with the stored context.

        :param labels: The labels to set.
        """
        for key, value in labels.items():
            self._context.set_label(key=key, value=value)

    def _collect_packagers(
        self, packagers: list[str], is_mandatory: bool, is_custom_packagers: bool
    ):
        """
        Collect packagers with the stored manager. The collection can ignore errors raised by setting the mandatory flag
        to False.

        :param packagers:           The list of packagers to collect.
        :param is_mandatory:        Whether the packagers are mandatory for the context run.
        :param is_custom_packagers: Whether the packagers to collect are user's custom or MLRun's builtins.
        """
        try:
            self._packagers_manager.collect_packagers(
                packagers=packagers,
                default_priority=self._CUSTOM_PACKAGERS_DEFAULT_PRIORITY
                if is_custom_packagers
                else self._BUILTIN_PACKAGERS_DEFAULT_PRIORITY,
            )
        except MLRunPackageCollectionError as error:
            if is_mandatory:
                raise error
            else:
                # If the packagers to collect were added manually by the user, the logger should write the collection
                # issue as a warning. Otherwise - for mlrun builtin packagers, a debug message will do.
                message = (
                    f"The given optional packagers '{packagers}' could not be imported due to the following error:\n"
                    f"'{error}'"
                )
                if is_custom_packagers:
                    self._context.logger.warn(message)
                else:
                    self._context.logger.debug(message)

    def _collect_mlrun_packagers(self):
        """
        Collect MLRun's builtin packagers. That include all mandatory packagers whom in charge of MLRun's requirements
        libraries, more optional commonly used libraries packagers and more `mlrun.frameworks` packagers. The priority
        will be as follows (from higher to lower priority):

        1. Optional `mlrun.frameworks` packagers
        2. MLRun's optional packagers
        3. MLRun's mandatory packagers (MLRun's requirements)
        """
        # Collect MLRun's requirements packagers (mandatory):
        self._collect_packagers(
            packagers=[
                f"mlrun.package.packagers.{module_name}_packagers.*"
                for module_name in self._MLRUN_REQUIREMENTS_PACKAGERS
            ],
            is_mandatory=True,
            is_custom_packagers=False,
        )

        # Add extra packagers for optional libraries:
        for module_name in self._EXTENDED_PACKAGERS:
            self._collect_packagers(
                packagers=[f"mlrun.package.packagers.{module_name}_packagers.*"],
                is_mandatory=False,
                is_custom_packagers=False,
            )

        # Add extra packagers from `mlrun.frameworks` package:
        for module_name in self._MLRUN_FRAMEWORKS_PACKAGERS:
            self._collect_packagers(
                packagers=[f"mlrun.frameworks.{module_name}.packagers.*"],
                is_mandatory=False,
                is_custom_packagers=False,
            )

    def _validate_objects_to_log_hints_length(
        self,
        outputs: list,
        log_hints: list[Union[dict[str, str], str, None]],
    ):
        """
        Validate the outputs and log hints are the same length. If they are not, warnings will be printed on what will
        be ignored.

        :param outputs:   List of outputs to log.
        :param log_hints: List of log hints (logging configurations) to use.
        """
        if len(outputs) != len(log_hints):
            self._context.logger.warn(
                f"The amount of outputs objects returned from the function ({len(outputs)}) does not match the amount "
                f"of provided log hints ({len(log_hints)})."
            )
            if len(outputs) > len(log_hints):
                ignored_outputs = [str(output) for output in outputs[len(log_hints) :]]
                self._context.logger.warn(
                    f"The following outputs will not be logged: {', '.join(ignored_outputs)}"
                )
            if len(outputs) < len(log_hints):
                ignored_log_hints = [
                    str(log_hint) for log_hint in log_hints[len(outputs) :]
                ]
                self._context.logger.warn(
                    f"The following log hints will be ignored: {', '.join(ignored_log_hints)}"
                )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import base64
import json
import typing
from urllib.parse import ParseResult, urlparse

import pydantic.v1
from deprecated import deprecated
from mergedeep import merge

import mlrun
import mlrun.errors

from ..secrets import get_secret_or_env


class DatastoreProfile(pydantic.v1.BaseModel):
    type: str
    name: str
    _private_attributes: list = ()

    class Config:
        extra = pydantic.v1.Extra.forbid

    @pydantic.v1.validator("name")
    @classmethod
    def lower_case(cls, v):
        return v.lower()

    @staticmethod
    def generate_secret_key(profile_name: str, project: str):
        secret_name_separator = "."
        full_key = (
            "datastore-profiles"
            + secret_name_separator
            + project
            + secret_name_separator
            + profile_name
        )
        return full_key

    def secrets(self) -> dict:
        return None

    def url(self, subpath) -> str:
        return None


class TemporaryClientDatastoreProfiles(metaclass=mlrun.utils.singleton.Singleton):
    def __init__(self):
        self._data = {}  # Initialize the dictionary

    def add(self, profile: DatastoreProfile):
        self._data[profile.name] = profile

    def get(self, key):
        return self._data.get(key, None)

    def remove(self, key):
        self._data.pop(key, None)


class DatastoreProfileBasic(DatastoreProfile):
    type: str = pydantic.v1.Field("basic")
    _private_attributes = "private"
    public: str
    private: typing.Optional[str] = None


class ConfigProfile(DatastoreProfile):
    """
    A profile class for managing configuration data with nested public and private attributes.
    This class extends DatastoreProfile to handle configuration settings, separating them into
    public and private dictionaries. Both dictionaries support nested structures, and the class
    provides functionality to merge these attributes when needed.

    Args:
        public (Optional[dict]): Dictionary containing public configuration settings,
            supporting nested structures
        private (Optional[dict]): Dictionary containing private/sensitive configuration settings,
            supporting nested structures

    Example:
        >>> public = {
            "database": {
                "host": "localhost",
                "port": 5432
            },
            "api_version": "v1"
        }
        >>> private = {
            "database": {
                "password": "secret123",
                "username": "admin"
            },
            "api_key": "xyz789"
        }
        >>> config = ConfigProfile("myconfig", public=public, private=private)

        # When attributes() is called, it merges public and private:
        # {
        #     "database": {
        #         "host": "localhost",
        #         "port": 5432,
        #         "password": "secret123",
        #         "username": "admin"
        #     },
        #     "api_version": "v1",
        #     "api_key": "xyz789"
        # }

    """

    type = "config"
    _private_attributes = "private"
    public: typing.Optional[dict] = None
    private: typing.Optional[dict] = None

    def attributes(self):
        res = {}
        if self.public:
            res = merge(res, self.public)
        if self.private:
            res = merge(res, self.private)
        return res


# TODO: Remove in 1.12.0
@deprecated(
    version="1.10.0",
    reason=(
        "This class is deprecated from mlrun 1.10.0, and will be removed in 1.12.0. "
        "Use `DatastoreProfileKafkaStream` instead."
    ),
    category=FutureWarning,
)
class DatastoreProfileKafkaTarget(DatastoreProfile):
    type: str = pydantic.v1.Field("kafka_target")
    _private_attributes = "kwargs_private"
    brokers: str
    topic: str
    kwargs_public: typing.Optional[dict]
    kwargs_private: typing.Optional[dict]

    def get_topic(self) -> typing.Optional[str]:
        return self.topic

    def attributes(self):
        attributes = {"brokers": self.brokers}
        if self.kwargs_public:
            attributes = merge(attributes, self.kwargs_public)
        if self.kwargs_private:
            attributes = merge(attributes, self.kwargs_private)
        return attributes


class DatastoreProfileKafkaStream(DatastoreProfile):
    type: str = pydantic.v1.Field("kafka_stream")
    _private_attributes = ("kwargs_private", "sasl_user", "sasl_pass")
    brokers: typing.Union[str, list[str]]
    topics: typing.Union[str, list[str]]
    group: typing.Optional[str] = "serving"
    initial_offset: typing.Optional[str] = "earliest"
    partitions: typing.Optional[typing.Union[str, list[str]]]
    sasl_user: typing.Optional[str]
    sasl_pass: typing.Optional[str]
    kwargs_public: typing.Optional[dict]
    kwargs_private: typing.Optional[dict]

    def get_topic(self) -> typing.Optional[str]:
        topics = [self.topics] if isinstance(self.topics, str) else self.topics
        return topics[0] if topics else None

    def attributes(self) -> dict[str, typing.Any]:
        attributes = {}
        if self.kwargs_public:
            attributes = merge(attributes, self.kwargs_public)
        if self.kwargs_private:
            attributes = merge(attributes, self.kwargs_private)

        topics = [self.topics] if isinstance(self.topics, str) else self.topics
        brokers = [self.brokers] if isinstance(self.brokers, str) else self.brokers

        attributes["brokers"] = brokers
        attributes["topics"] = topics
        attributes["group"] = self.group
        attributes["initial_offset"] = self.initial_offset
        if self.partitions is not None:
            attributes["partitions"] = self.partitions
        if sasl := mlrun.datastore.utils.KafkaParameters(attributes).sasl(
            usr=self.sasl_user, pwd=self.sasl_pass
        ):
            attributes["sasl"] = sasl
        return attributes


# TODO: Remove in 1.12.0
@deprecated(
    version="1.10.0",
    reason=(
        "This class is deprecated from mlrun 1.10.0, and will be removed in 1.12.0. "
        "Use `DatastoreProfileKafkaStream` instead."
    ),
    category=FutureWarning,
)
class DatastoreProfileKafkaSource(DatastoreProfileKafkaStream):
    type: str = pydantic.v1.Field("kafka_source")


class DatastoreProfileV3io(DatastoreProfile):
    type: str = pydantic.v1.Field("v3io")
    v3io_access_key: typing.Optional[str] = None
    _private_attributes = "v3io_access_key"

    def url(self, subpath):
        subpath = subpath.lstrip("/")
        return f"v3io:///{subpath}"

    def secrets(self) -> dict:
        res = {}
        if self.v3io_access_key:
            res["V3IO_ACCESS_KEY"] = self.v3io_access_key
        return res


class DatastoreProfileS3(DatastoreProfile):
    type: str = pydantic.v1.Field("s3")
    _private_attributes = ("access_key_id", "secret_key")
    endpoint_url: typing.Optional[str] = None
    force_non_anonymous: typing.Optional[str] = None
    profile_name: typing.Optional[str] = None
    assume_role_arn: typing.Optional[str] = None
    access_key_id: typing.Optional[str] = None
    secret_key: typing.Optional[str] = None
    bucket: str

    def secrets(self) -> dict:
        res = {}
        if self.access_key_id:
            res["AWS_ACCESS_KEY_ID"] = self.access_key_id
        if self.secret_key:
            res["AWS_SECRET_ACCESS_KEY"] = self.secret_key
        if self.endpoint_url:
            res["AWS_ENDPOINT_URL_S3"] = self.endpoint_url
        if self.force_non_anonymous:
            res["S3_NON_ANONYMOUS"] = self.force_non_anonymous
        if self.profile_name:
            res["AWS_PROFILE"] = self.profile_name
        if self.assume_role_arn:
            res["MLRUN_AWS_ROLE_ARN"] = self.assume_role_arn
        return res

    def url(self, subpath):
        # TODO: There is an inconsistency with DatastoreProfileGCS. In DatastoreProfileGCS,
        # we assume that the subpath can begin without a '/' character,
        # while here we assume it always starts with one.
        if self.bucket:
            return f"s3://{self.bucket}{subpath}"
        else:
            return f"s3:/{subpath}"


class DatastoreProfileRedis(DatastoreProfile):
    type: str = pydantic.v1.Field("redis")
    _private_attributes = ("username", "password")
    endpoint_url: str
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None

    def url_with_credentials(self):
        parsed_url = urlparse(self.endpoint_url)
        username = self.username
        password = self.password
        netloc = parsed_url.hostname
        if username:
            if password:
                netloc = f"{username}:{password}@{parsed_url.hostname}"
            else:
                netloc = f"{username}@{parsed_url.hostname}"

        if parsed_url.port:
            netloc += f":{parsed_url.port}"

        new_parsed_url = ParseResult(
            scheme=parsed_url.scheme,
            netloc=netloc,
            path=parsed_url.path,
            params=parsed_url.params,
            query=parsed_url.query,
            fragment=parsed_url.fragment,
        )
        return new_parsed_url.geturl()

    def secrets(self) -> dict:
        res = {}
        if self.username:
            res["REDIS_USER"] = self.username
        if self.password:
            res["REDIS_PASSWORD"] = self.password
        return res

    def url(self, subpath):
        return self.endpoint_url + subpath


class DatastoreProfileDBFS(DatastoreProfile):
    type: str = pydantic.v1.Field("dbfs")
    _private_attributes = ("token",)
    endpoint_url: typing.Optional[str] = None  # host
    token: typing.Optional[str] = None

    def url(self, subpath) -> str:
        return f"dbfs://{subpath}"

    def secrets(self) -> dict:
        res = {}
        if self.token:
            res["DATABRICKS_TOKEN"] = self.token
        if self.endpoint_url:
            res["DATABRICKS_HOST"] = self.endpoint_url
        return res


class DatastoreProfileGCS(DatastoreProfile):
    type: str = pydantic.v1.Field("gcs")
    _private_attributes = ("gcp_credentials",)
    credentials_path: typing.Optional[str] = None  # path to file.
    gcp_credentials: typing.Optional[typing.Union[str, dict]] = None
    bucket: str

    @pydantic.v1.validator("gcp_credentials", pre=True, always=True)
    @classmethod
    def convert_dict_to_json(cls, v):
        if isinstance(v, dict):
            return json.dumps(v)
        return v

    def url(self, subpath) -> str:
        # TODO: but there's something wrong with the subpath being assumed to not start with a slash here,
        # but the opposite assumption is made in S3.
        if subpath.startswith("/"):
            #  in gcs the path after schema is starts with bucket, wherefore it should not start with "/".
            subpath = subpath[1:]
        if self.bucket:
            return (
                f"gcs://{self.bucket}/{subpath}" if subpath else f"gcs://{self.bucket}"
            )
        else:
            return f"gcs://{subpath}"

    def secrets(self) -> dict:
        res = {}
        if self.credentials_path:
            res["GOOGLE_APPLICATION_CREDENTIALS"] = self.credentials_path
        if self.gcp_credentials:
            res["GCP_CREDENTIALS"] = self.gcp_credentials
        return res


class DatastoreProfileAzureBlob(DatastoreProfile):
    type: str = pydantic.v1.Field("az")
    _private_attributes = (
        "connection_string",
        "account_key",
        "client_secret",
        "sas_token",
        "credential",
    )
    connection_string: typing.Optional[str] = None
    account_name: typing.Optional[str] = None
    account_key: typing.Optional[str] = None
    tenant_id: typing.Optional[str] = None
    client_id: typing.Optional[str] = None
    client_secret: typing.Optional[str] = None
    sas_token: typing.Optional[str] = None
    credential: typing.Optional[str] = None
    container: str

    def url(self, subpath) -> str:
        if subpath.startswith("/"):
            #  in azure the path after schema is starts with container, wherefore it should not start with "/".
            subpath = subpath[1:]
        if self.container:
            return (
                f"az://{self.container}/{subpath}"
                if subpath
                else f"az://{self.container}"
            )
        else:
            return f"az://{subpath}"

    def secrets(self) -> dict:
        res = {}
        if self.connection_string:
            res["connection_string"] = self.connection_string
        if self.account_name:
            res["account_name"] = self.account_name
        if self.account_key:
            res["account_key"] = self.account_key
        if self.tenant_id:
            res["tenant_id"] = self.tenant_id
        if self.client_id:
            res["client_id"] = self.client_id
        if self.client_secret:
            res["client_secret"] = self.client_secret
        if self.sas_token:
            res["sas_token"] = self.sas_token
        if self.credential:
            res["credential"] = self.credential
        return res


class DatastoreProfileHdfs(DatastoreProfile):
    type: str = pydantic.v1.Field("hdfs")
    _private_attributes = "token"
    host: typing.Optional[str] = None
    port: typing.Optional[int] = None
    http_port: typing.Optional[int] = None
    user: typing.Optional[str] = None

    def secrets(self) -> dict:
        res = {}
        if self.host:
            res["HDFS_HOST"] = self.host
        if self.port:
            res["HDFS_PORT"] = self.port
        if self.port:
            res["HDFS_HTTP_PORT"] = self.http_port
        if self.user:
            res["HDFS_USER"] = self.user
        return res or None

    def url(self, subpath):
        return f"webhdfs://{self.host}:{self.http_port}{subpath}"


class DatastoreProfileTDEngine(DatastoreProfile):
    """
    A profile that holds the required parameters for a TDEngine database, with the websocket scheme.
    https://docs.tdengine.com/developer-guide/connecting-to-tdengine/#websocket-connection
    """

    type: str = pydantic.v1.Field("taosws")
    _private_attributes = ["password"]
    user: str
    # The password cannot be empty in real world scenarios. It's here just because of the profiles completion design.
    password: typing.Optional[str]
    host: str
    port: int

    def dsn(self) -> str:
        """Get the Data Source Name of the configured TDEngine profile."""
        return f"{self.type}://{self.user}:{self.password}@{self.host}:{self.port}"

    @classmethod
    def from_dsn(cls, dsn: str, profile_name: str) -> "DatastoreProfileTDEngine":
        """
        Construct a TDEngine profile from DSN (connection string) and a name for the profile.

        :param dsn:          The DSN (Data Source Name) of the TDEngine database, e.g.: ``"taosws://root:taosdata@localhost:6041"``.
        :param profile_name: The new profile's name.
        :return:             The TDEngine profile.
        """
        parsed_url = urlparse(dsn)
        return cls(
            name=profile_name,
            user=parsed_url.username,
            password=parsed_url.password,
            host=parsed_url.hostname,
            port=parsed_url.port,
        )


class OpenAIProfile(DatastoreProfile):
    type: str = pydantic.v1.Field("openai")
    _private_attributes = "api_key"
    api_key: typing.Optional[str] = None
    organization: typing.Optional[str] = None
    project: typing.Optional[str] = None
    base_url: typing.Optional[str] = None
    timeout: typing.Optional[float] = None
    max_retries: typing.Optional[int] = None

    def secrets(self) -> dict:
        res = {}
        if self.api_key:
            res["OPENAI_API_KEY"] = self.api_key
        if self.organization:
            res["OPENAI_ORG_ID"] = self.organization
        if self.project:
            res["OPENAI_PROJECT_ID"] = self.project
        if self.base_url:
            res["OPENAI_BASE_URL"] = self.base_url
        if self.timeout:
            res["OPENAI_TIMEOUT"] = self.timeout
        if self.max_retries:
            res["OPENAI_MAX_RETRIES"] = self.max_retries
        return res

    def url(self, subpath):
        return f"{self.type}://{subpath.lstrip('/')}"


class HuggingFaceProfile(DatastoreProfile):
    type: str = pydantic.v1.Field("huggingface")
    _private_attributes = ("token", "model_kwargs")
    task: typing.Optional[str] = None
    token: typing.Optional[str] = None
    device: typing.Optional[typing.Union[int, str]] = None
    device_map: typing.Union[str, dict[str, typing.Union[int, str]], None] = None
    trust_remote_code: bool = None
    model_kwargs: typing.Optional[dict[str, typing.Any]] = None

    def secrets(self) -> dict:
        keys = {
            "HF_TASK": self.task,
            "HF_TOKEN": self.token,
            "HF_DEVICE": self.device,
            "HF_DEVICE_MAP": self.device_map,
            "HF_TRUST_REMOTE_CODE": self.trust_remote_code,
            "HF_MODEL_KWARGS": self.model_kwargs,
        }
        return {k: v for k, v in keys.items() if v}

    def url(self, subpath):
        return f"{self.type}://{subpath.lstrip('/')}"


_DATASTORE_TYPE_TO_PROFILE_CLASS: dict[str, type[DatastoreProfile]] = {
    "v3io": DatastoreProfileV3io,
    "s3": DatastoreProfileS3,
    "redis": DatastoreProfileRedis,
    "basic": DatastoreProfileBasic,
    "kafka_target": DatastoreProfileKafkaTarget,
    "kafka_source": DatastoreProfileKafkaSource,
    "kafka_stream": DatastoreProfileKafkaStream,
    "dbfs": DatastoreProfileDBFS,
    "gcs": DatastoreProfileGCS,
    "az": DatastoreProfileAzureBlob,
    "hdfs": DatastoreProfileHdfs,
    "taosws": DatastoreProfileTDEngine,
    "config": ConfigProfile,
    "openai": OpenAIProfile,
    "huggingface": HuggingFaceProfile,
}


class DatastoreProfile2Json(pydantic.v1.BaseModel):
    @staticmethod
    def _to_json(attributes):
        # First, base64 encode the values
        encoded_dict = {
            k: base64.b64encode(str(v).encode()).decode() for k, v in attributes.items()
        }
        # Then, return the dictionary as a JSON string with no spaces
        return json.dumps(encoded_dict).replace(" ", "")

    @staticmethod
    def get_json_public(profile: DatastoreProfile) -> str:
        return DatastoreProfile2Json._to_json(
            {
                k: v
                for k, v in profile.dict().items()
                if str(k) not in profile._private_attributes
            }
        )

    @staticmethod
    def get_json_private(profile: DatastoreProfile) -> str:
        return DatastoreProfile2Json._to_json(
            {
                k: v
                for k, v in profile.dict().items()
                if str(k) in profile._private_attributes
            }
        )

    @staticmethod
    def create_from_json(public_json: str, private_json: str = "{}"):
        attributes = json.loads(public_json)
        attributes_public = {
            k: base64.b64decode(str(v).encode()).decode() for k, v in attributes.items()
        }
        attributes = json.loads(private_json)
        attributes_private = {
            k: base64.b64decode(str(v).encode()).decode() for k, v in attributes.items()
        }
        decoded_dict = merge(attributes_public, attributes_private)

        def safe_literal_eval(value):
            try:
                return ast.literal_eval(value)
            except (ValueError, SyntaxError):
                return value

        decoded_dict = {k: safe_literal_eval(v) for k, v in decoded_dict.items()}
        datastore_type = decoded_dict.get("type")
        ds_profile_factory = _DATASTORE_TYPE_TO_PROFILE_CLASS
        if datastore_type in ds_profile_factory:
            return ds_profile_factory[datastore_type].parse_obj(decoded_dict)
        else:
            if datastore_type:
                reason = f"unexpected type '{decoded_dict['type']}'"
            else:
                reason = "missing type"
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Datastore profile failed to create from json due to {reason}"
            )


def datastore_profile_read(url, project_name="", secrets: typing.Optional[dict] = None):
    """
    Read and retrieve a datastore profile from a given URL.

    This function retrieves a datastore profile either from temporary client storage,
    or from the MLRun database. It handles both client-side and server-side profile formats
    and performs necessary conversions.

    Args:
        url (str): A URL with 'ds' scheme pointing to the datastore profile
            (e.g., 'ds://profile-name').
        project_name (str, optional): The project name where the profile is stored.
            Defaults to MLRun's active project.
        secrets (dict, optional): Dictionary containing secrets needed for profile retrieval.

    Returns:
        DatastoreProfile: The retrieved datastore profile object.

    Raises:
        MLRunInvalidArgumentError: In the following cases:
            - If the URL scheme is not 'ds'
            - If the profile cannot be retrieved from either server or local environment

    Note:
       When running from a client environment (outside MLRun pods), private profile information
       is not accessible. In this case, use register_temporary_client_datastore_profile() to
       register the profile with credentials for your local session. When running inside MLRun
       pods, the private information is automatically available and no temporary registration is needed.
    """

    parsed_url = urlparse(url)
    if parsed_url.scheme.lower() != "ds":
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"resource URL '{url}' cannot be read as a datastore profile because its scheme is not 'ds'"
        )

    profile_name = parsed_url.hostname
    project_name = project_name or mlrun.mlconf.active_project
    datastore = TemporaryClientDatastoreProfiles().get(profile_name)
    if datastore:
        return datastore
    public_profile = mlrun.db.get_run_db().get_datastore_profile(
        profile_name, project_name
    )
    # The mlrun.db.get_run_db().get_datastore_profile() function is capable of returning
    # two distinct types of objects based on its execution context.
    # If it operates from the client or within the pod (which is the common scenario),
    # it yields an instance of `mlrun.datastore.DatastoreProfile`. Conversely,
    # when executed on the server with a direct call to `sqldb`, it produces an instance of
    # mlrun.common.schemas.DatastoreProfile.
    # In the latter scenario, an extra conversion step is required to transform the object
    # into mlrun.datastore.DatastoreProfile.
    if isinstance(public_profile, mlrun.common.schemas.DatastoreProfile):
        public_profile = DatastoreProfile2Json.create_from_json(
            public_json=public_profile.object
        )
    project_ds_name_private = DatastoreProfile.generate_secret_key(
        profile_name, project_name
    )
    private_body = get_secret_or_env(project_ds_name_private, secret_provider=secrets)
    if not public_profile or not private_body:
        raise mlrun.errors.MLRunNotFoundError(
            f"Unable to retrieve the datastore profile '{url}' from either the server or local environment. "
            "Make sure the profile is registered correctly, or if running in a local environment, "
            "use register_temporary_client_datastore_profile() to provide credentials locally."
        )

    datastore = DatastoreProfile2Json.create_from_json(
        public_json=DatastoreProfile2Json.get_json_public(public_profile),
        private_json=private_body,
    )
    return datastore


def register_temporary_client_datastore_profile(profile: DatastoreProfile):
    """Register the datastore profile.
    This profile is temporary and remains valid only for the duration of the caller's session.
    It's beneficial for testing purposes.
    """
    TemporaryClientDatastoreProfiles().add(profile)


def remove_temporary_client_datastore_profile(profile_name: str):
    TemporaryClientDatastoreProfiles().remove(profile_name)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO, StringIO

import pandas as pd

import mlrun
import mlrun.utils.helpers

from .base import DataStore, FileStats


class InMemoryStore(DataStore):
    def __init__(self):
        super().__init__(None, "memory", "memory", "")
        self._items = {}

    @property
    def url(self):
        return "memory://"

    def _get_parent_secret(self, key):
        return None

    def _get_item(self, key):
        if key not in self._items:
            raise mlrun.errors.MLRunNotFoundError(
                f"item {key} not found in memory store"
            )
        return self._items[key]

    def get(self, key, size=None, offset=0):
        item = self._get_item(key)
        return item

    def put(self, key, data, append=False):
        if append and key in self._items:
            self._items[key] = self._items[key] + data
        else:
            self._items[key] = data

    def upload(self, key, src_path):
        with open(src_path, "rb") as fp:
            self._items[key] = fp.read()

    def stat(self, key):
        return FileStats(size=len(self._get_item(key)), modified=0)

    def listdir(self, key):
        return []

    def as_df(self, url, subpath, columns=None, df_module=None, format="", **kwargs):
        item = self._get_item(subpath)
        if hasattr(item, "to_csv"):  # detect if it is a dataframe type
            return item
        if isinstance(item, str):
            item = StringIO(item)
        else:
            item = BytesIO(item)

        df_module = df_module or pd
        if url.endswith(".csv") or format == "csv":
            if columns:
                kwargs["usecols"] = columns
            reader = df_module.read_csv
        elif mlrun.utils.helpers.is_parquet_file(url, format):
            if columns:
                kwargs["columns"] = columns
            reader = df_module.read_parquet
        elif url.endswith(".json") or format == "json":
            reader = df_module.read_json
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(f"file type unhandled {url}")
        # InMemoryStore store – don't pass filters
        for field in ["time_column", "start_time", "end_time", "additional_filters"]:
            kwargs.pop(field, None)

        return reader(item, **kwargs)

    def rm(self, path, recursive=False, maxdepth=None):
        self._items.pop(path, None)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from .._common import ModelHandler


class DLModelHandler(ModelHandler, ABC):
    """
    Abstract class for a deep learning framework model handling, enabling loading, saving and logging it during runs.
    """

    # Constant artifact names:
    _WEIGHTS_FILE_ARTIFACT_NAME = "{}_weights_file"

    def _get_weights_file_artifact_name(self) -> str:
        """
        Get the standard name for the weights file artifact.

        :return: The weights file artifact name.
        """
        return self._WEIGHTS_FILE_ARTIFACT_NAME.format(self._model_name)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import mlrun.errors

from .dask_merger import DaskFeatureMerger
from .job import RemoteVectorResponse, run_merge_job  # noqa
from .local_merger import LocalFeatureMerger
from .spark_merger import SparkFeatureMerger
from .storey_merger import StoreyFeatureMerger

mergers = {
    "local": LocalFeatureMerger,
    "dask": DaskFeatureMerger,
    "spark": SparkFeatureMerger,
    "storey": StoreyFeatureMerger,
}


def get_merger(kind):
    if not kind:
        return LocalFeatureMerger
    merger = mergers.get(kind)
    if not merger:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"No merger was found for engine '{kind}'. Supported engines: {', '.join(mergers)}."
        )
    return merger



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from collections import OrderedDict
from typing import Callable, Optional, Union

from ..config import config
from .context_handler import ContextHandler
from .errors import (
    MLRunPackageCollectionError,
    MLRunPackageError,
    MLRunPackagePackingError,
    MLRunPackageUnpackingError,
)
from .packager import Packager
from .packagers import DefaultPackager
from .packagers_manager import PackagersManager
from .utils import (
    ArchiveSupportedFormat,
    ArtifactType,
    LogHintKey,
    StructFileSupportedFormat,
)


def handler(
    labels: Optional[dict[str, str]] = None,
    outputs: Optional[list[Union[str, dict[str, str]]]] = None,
    inputs: Union[bool, dict[str, Union[str, type]]] = True,
):
    """
    MLRun's handler is a decorator to wrap a function and enable setting labels, parsing inputs (`mlrun.DataItem`) using
    type hints and log returning outputs using log hints.

    Notice: this decorator is now appplied automatically with the release of `mlrun.package`. It should not be used
    manually.

    :param labels:  Labels to add to the run. Expecting a dictionary with the labels names as keys. Default: None.
    :param outputs: Log hints (logging configurations) for the function's returned values. Expecting a list of the
                    following values:

                    * `str` - A string in the format of '{key}:{artifact_type}'. If a string was given without ':' it
                      will indicate the key, and the artifact type will be according to the returned value type's
                      default artifact type. The artifact types supported are listed in the relevant type packager.
                    * `dict[str, str]` - A dictionary of logging configuration. the key 'key' is mandatory for the
                      logged artifact key.
                    * None - Do not log the output.

                    If the list length is not equal to the total amount of returned values from the function, those
                    without log hints will be ignored.

                    Default: None - meaning no outputs will be logged.

    :param inputs: Type hints (parsing configurations) for the arguments passed as inputs via the `run` method of an
                   MLRun function. Can be passed as a boolean value or a dictionary:

                   * True - Parse all found inputs to the assigned type hint in the function's signature. If there is no
                     type hint assigned, the value will remain an `mlrun.DataItem`.
                   * False - Do not parse inputs, leaving the inputs as `mlrun.DataItem`.
                   * dict[str, Union[Type, str]] - A dictionary with argument name as key and the expected type to parse
                     the `mlrun.DataItem` to. The expected type can be a string as well, idicating the full module path.

                   Default: True - meaning inputs will be parsed from DataItem's as long as they are type hinted.

    Example::

            import mlrun

            @mlrun.handler(
                outputs=[
                    "my_string",
                    None,
                    {"key": "my_array", "artifact_type": "file", "file_format": "npy"},
                    "my_multiplier: reuslt"
                ]
            )
            def my_handler(array: np.ndarray, m: int):
                m += 1
                array = array * m
                return "I will be logged", "I won't be logged", array, m

            >>> mlrun_function = mlrun.code_to_function("my_code.py", kind="job")
            >>> run_object = mlrun_function.run(
            ...     handler="my_handler",
            ...     inputs={"array": "store://my_array_Artifact"},
            ...     params={"m": 2}
            ... )
            >>> run_object.outputs
            {'my_string': 'I will be logged', 'my_array': 'store://...', 'my_multiplier': 3}
    """

    def decorator(func: Callable):
        def wrapper(*args: tuple, **kwargs: dict):
            nonlocal labels
            nonlocal outputs
            nonlocal inputs

            # Set default `inputs` - inspect the full signature and add the user's input on top of it:
            if inputs:
                # Get the available parameters type hints from the function's signature:
                func_signature = inspect.signature(func)
                parameters = OrderedDict(
                    {
                        parameter.name: parameter.annotation
                        for parameter in func_signature.parameters.values()
                    }
                )
                # If user input is given, add it on top of the collected defaults (from signature):
                if isinstance(inputs, dict):
                    parameters.update(inputs)
                inputs = parameters

            # Create a context handler and look for a context:
            cxt_handler = ContextHandler()
            cxt_handler.look_for_context(args=args, kwargs=kwargs)

            # If an MLRun context is found, parse arguments pre-run (kwargs are parsed inplace):
            if cxt_handler.is_context_available() and inputs:
                args = cxt_handler.parse_inputs(
                    args=args, kwargs=kwargs, type_hints=inputs
                )

            # Call the original function and get the returning values:
            func_outputs = func(*args, **kwargs)

            # If an MLRun context is found, set the given labels and log the returning values to MLRun via the context:
            if cxt_handler.is_context_available():
                if labels:
                    # TODO: Should deprecate this labels
                    cxt_handler.set_labels(labels=labels)
                if outputs:
                    cxt_handler.log_outputs(
                        outputs=func_outputs
                        if type(func_outputs) is tuple
                        and not config.packagers.pack_tuples
                        else [func_outputs],
                        log_hints=outputs,
                    )
                    return  # Do not return any values as the returning values were logged to MLRun.
            return func_outputs

        # Make sure to pass the wrapped function's signature (argument list, type hints and doc strings) to the wrapper:
        wrapper = functools.wraps(func)(wrapper)

        return wrapper

    return decorator



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import typing
from http import HTTPStatus

import aiohttp
import requests


class MLRunBaseError(Exception):
    """
    A base class from which all other exceptions inherit.
    If you want to catch all errors that the MLRun SDK might raise,
    catch this base exception.
    """

    pass


class MLRunTaskNotReadyError(MLRunBaseError):
    """indicate we are trying to read a value which is not ready
    or need to come from a job which is in progress"""


MLRunTaskNotReady = MLRunTaskNotReadyError  # kept for BC only


class MLRunHTTPError(MLRunBaseError, requests.HTTPError):
    def __init__(
        self,
        *args,
        response: typing.Optional[
            typing.Union[requests.Response, aiohttp.ClientResponse]
        ] = None,
        status_code: typing.Optional[int] = None,
        **kwargs,
    ):
        # because response object is probably with an error, it returns False, so we
        # should use 'is None' specifically
        if response is None:
            response = requests.Response()
        if status_code:
            response.status_code = status_code

        if isinstance(response, aiohttp.ClientResponse):
            if "request" not in kwargs:
                kwargs["request"] = response.request_info

            # consolidate the response object to be a requests.Response object-like
            setattr(
                response, "status_code", status_code if status_code else response.status
            )

        requests.HTTPError.__init__(self, *args, response=response, **kwargs)


class MLRunHTTPStatusError(MLRunHTTPError):
    """
    When an error has a matching http status code it is "HTTP statusable"
    HTTP Status errors should inherit from this class and set the right status code in the
    error_status_code attribute
    """

    error_status_code = None

    def __init__(self, *args, response: requests.Response = None, **kwargs):
        super().__init__(
            *args, response=response, status_code=self.error_status_code, **kwargs
        )


def raise_for_status(
    response: typing.Union[
        requests.Response,
        aiohttp.ClientResponse,
    ],
    message: typing.Optional[str] = None,
):
    """
    Raise a specific MLRunSDK error depending on the given response status code.
    If no specific error exists, raises an MLRunHTTPError
    """
    try:
        response.raise_for_status()
    except (requests.HTTPError, aiohttp.ClientResponseError) as exc:
        error_message = err_to_str(exc) if not message else message
        status_code = (
            response.status_code
            if hasattr(response, "status_code")
            else response.status
        )
        try:
            raise STATUS_ERRORS[status_code](error_message, response=response) from exc
        except KeyError:
            raise MLRunHTTPError(error_message, response=response) from exc


def err_for_status_code(status_code: int, message: typing.Optional[str] = None):
    """
    Return a specific MLRunSDK error depending on the given response status code.
    If no specific error exists, returns an MLRunHTTPError.
    Usage example:
    >>> try:
    >>>     ...
    >>> except ExcWithStatusCode as exc:
    >>>     raise err_for_status_code(exc.status_code, exc.message) from exc
    """
    try:
        return STATUS_ERRORS[int(status_code)](message)
    except KeyError:
        return MLRunHTTPError(message)


def err_to_str(err):
    if not err:
        return ""

    if isinstance(err, str):
        return err

    errors = []
    error_strings = []
    while err and err not in errors:
        errors.append(err)
        err_msg = str(err)
        if not err_msg:
            err_msg = repr(err)
        error_strings.append(err_msg)
        err = err.__cause__

    err_msg = ", caused by: ".join(error_strings)

    # in case the error string is longer than 32k, we truncate it
    # the truncation takes the first 16k, then the last 16k characters
    if len(err_msg) > 32_000:
        err_msg = err_msg[:16_000] + "...truncated..." + err_msg[-16_000:]
    return err_msg


# Specific Errors
class MLRunUnauthorizedError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.UNAUTHORIZED.value


class MLRunAccessDeniedError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.FORBIDDEN.value


class MLRunNotFoundError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.NOT_FOUND.value


class MLRunPaginationEndOfResultsError(MLRunNotFoundError):
    pass


class MLRunBadRequestError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunInvalidArgumentError(MLRunHTTPStatusError, ValueError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunModelLimitExceededError(MLRunHTTPStatusError, ValueError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunInvalidArgumentTypeError(MLRunHTTPStatusError, TypeError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunConflictError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.CONFLICT.value


class MLRunPreconditionFailedError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.PRECONDITION_FAILED.value


class MLRunIncompatibleVersionError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunInternalServerError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value


class MLRunNotImplementedServerError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.NOT_IMPLEMENTED.value


class MLRunServiceUnavailableError(MLRunHTTPStatusError):
    error_status_code = HTTPStatus.SERVICE_UNAVAILABLE.value


class MLRunRuntimeError(MLRunHTTPStatusError, RuntimeError):
    error_status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value


class MLRunMissingDependencyError(MLRunInternalServerError):
    pass


class MLRunTimeoutError(MLRunHTTPStatusError, TimeoutError):
    error_status_code = HTTPStatus.GATEWAY_TIMEOUT.value


class MLRunInvalidMMStoreTypeError(MLRunHTTPStatusError, ValueError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunStreamConnectionFailureError(MLRunHTTPStatusError, ValueError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunTSDBConnectionFailureError(MLRunHTTPStatusError, ValueError):
    error_status_code = HTTPStatus.BAD_REQUEST.value


class MLRunMissingProjectError(MLRunBadRequestError):
    default_message = "Project must be provided"

    def __init__(self, message=None):
        super().__init__(message or self.default_message)


class MLRunRetryExhaustedError(Exception):
    pass


class MLRunTaskCancelledError(Exception):
    pass


class MLRunValueError(ValueError):
    pass


class MLRunFatalFailureError(Exception):
    """
    Internal exception meant to be used inside mlrun.utils.helpers.retry_until_successful to signal the loop not to
    retry
    Allowing to pass to original exception that will be raised from the loop (instead of this exception)
    """

    def __init__(
        self, *args, original_exception: typing.Optional[Exception] = None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.original_exception = original_exception


class ModelRunnerError(MLRunBaseError):
    def __init__(self, models_errors: dict[str:str], *args) -> None:
        self.models_errors = models_errors
        super().__init__(self.__repr__(), *args)

    def __repr__(self):
        return f"ModelRunnerError: {repr(self.models_errors)}"

    def __copy__(self):
        return type(self)(models_errors=self.models_errors)

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(self.models_errors, memo))

    def get_errors(self):
        return self.models_errors

    def get_model_error(self, model: str):
        return self.models_errors.get(model)


STATUS_ERRORS = {
    HTTPStatus.BAD_REQUEST.value: MLRunBadRequestError,
    HTTPStatus.UNAUTHORIZED.value: MLRunUnauthorizedError,
    HTTPStatus.FORBIDDEN.value: MLRunAccessDeniedError,
    HTTPStatus.NOT_FOUND.value: MLRunNotFoundError,
    HTTPStatus.CONFLICT.value: MLRunConflictError,
    HTTPStatus.PRECONDITION_FAILED.value: MLRunPreconditionFailedError,
    HTTPStatus.INTERNAL_SERVER_ERROR.value: MLRunInternalServerError,
    HTTPStatus.SERVICE_UNAVAILABLE.value: MLRunServiceUnavailableError,
    HTTPStatus.NOT_IMPLEMENTED.value: MLRunNotImplementedServerError,
}

EXPECTED_ERRORS = (MLRunPaginationEndOfResultsError,)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
from copy import copy
from typing import Union

import numpy as np

import mlrun
from mlrun.feature_store import FeatureSet

from ..model import ModelObj, ObjectList

# Forward reference for type annotations


class _JoinStep(ModelObj):
    def __init__(
        self,
        name: typing.Optional[str] = None,
        left_step_name: typing.Optional[str] = None,
        right_step_name: typing.Optional[str] = None,
        left_feature_set_names: typing.Optional[Union[str, list[str]]] = None,
        right_feature_set_name: typing.Optional[str] = None,
        join_type: str = "inner",
        asof_join: bool = False,
    ):
        self.name = name
        self.left_step_name = left_step_name
        self.right_step_name = right_step_name
        self.left_feature_set_names = (
            left_feature_set_names
            if left_feature_set_names is None
            or isinstance(left_feature_set_names, list)
            else [left_feature_set_names]
        )
        self.right_feature_set_name = right_feature_set_name
        self.join_type = join_type
        self.asof_join = asof_join

        self.left_keys = []
        self.right_keys = []

    def init_join_keys(
        self,
        feature_set_objects: ObjectList,
        vector,
        entity_rows_keys: typing.Optional[list[str]] = None,
    ):
        if feature_set_objects[self.right_feature_set_name].is_connectable_to_df(
            entity_rows_keys
        ):
            self.left_keys, self.right_keys = [
                list(
                    feature_set_objects[
                        self.right_feature_set_name
                    ].spec.entities.keys()
                )
            ] * 2

        if (
            self.join_type == JoinGraph.first_join_type
            or not self.left_feature_set_names
        ):
            self.join_type = (
                "inner"
                if self.join_type == JoinGraph.first_join_type
                else self.join_type
            )
            return

        for left_fset in self.left_feature_set_names:
            current_left_keys = feature_set_objects[left_fset].extract_relation_keys(
                feature_set_objects[self.right_feature_set_name],
                vector.get_feature_set_relations(feature_set_objects[left_fset]),
            )
            current_right_keys = list(
                feature_set_objects[self.right_feature_set_name].spec.entities.keys()
            )
            for i in range(len(current_left_keys)):
                if (
                    current_left_keys[i] not in self.left_keys
                    and current_right_keys[i] not in self.right_keys
                ):
                    self.left_keys.append(current_left_keys[i])
                    self.right_keys.append(current_right_keys[i])

        if not self.left_keys:
            raise mlrun.errors.MLRunRuntimeError(
                f"{self.name} can't be preform due to undefined relation between "
                f"{self.left_feature_set_names} to {self.right_feature_set_name}"
            )


class JoinGraph(ModelObj):
    """
    A class that represents a graph of data joins between feature sets
    """

    default_graph_name = "$__join_graph_fv__$"
    first_join_type = "first"
    _dict_fields = ["name", "first_feature_set", "steps"]

    def __init__(
        self,
        name: typing.Optional[str] = None,
        first_feature_set: Union[str, FeatureSet] = None,
    ):
        """
        JoinGraph is a class that represents a graph of data joins between feature sets. It allows users to define
        data joins step by step, specifying the join type for each step. The graph can be used to build a sequence of
        joins that will be executed in order, allowing the creation of complex join operations between feature sets.


        Example:
        # Create a new JoinGraph and add steps for joining feature sets.
        join_graph = JoinGraph(name="my_join_graph", first_feature_set="featureset1")
        join_graph.inner("featureset2")
        join_graph.left("featureset3", asof_join=True)


        :param name:                (str, optional) The name of the join graph. If not provided,
                                    a default name will be used.
        :param first_feature_set:   (str or FeatureSet, optional) The first feature set to join. It can be
                                    specified either as a string representing the name of the feature set or as a
                                    FeatureSet object.
        """
        self.name = name or self.default_graph_name
        self._steps: ObjectList = None
        self._feature_sets = None
        if first_feature_set:
            self._start(first_feature_set)

    def inner(self, other_operand: typing.Union[str, FeatureSet]):
        """
        Specifies an inner join with the given feature set

        :param other_operand:       (str or FeatureSet) The name of the feature set or a FeatureSet object to join with.

        :return:                    JoinGraph: The updated JoinGraph object with the specified inner join.
        """
        return self._join_operands(other_operand, "inner")

    def outer(self, other_operand: typing.Union[str, FeatureSet]):
        """
        Specifies an outer join with the given feature set

        :param other_operand:       (str or FeatureSet) The name of the feature set or a FeatureSet object to join with.
        :return:                    JoinGraph: The updated JoinGraph object with the specified outer join.
        """
        return self._join_operands(other_operand, "outer")

    def left(self, other_operand: typing.Union[str, FeatureSet], asof_join):
        """
        Specifies a left join with the given feature set

        :param other_operand:       (str or FeatureSet) The name of the feature set or a FeatureSet object to join with.
        :param asof_join:           (bool) A flag indicating whether to perform an as-of join.

        :return:                    JoinGraph: The updated JoinGraph object with the specified left join.
        """
        return self._join_operands(other_operand, "left", asof_join=asof_join)

    def right(self, other_operand: typing.Union[str, FeatureSet]):
        """
        Specifies a right join with the given feature set

        :param other_operand:       (str or FeatureSet) The name of the feature set or a FeatureSet object to join with.

        :return:                    JoinGraph: The updated JoinGraph object with the specified right join.
        """
        return self._join_operands(other_operand, "right")

    def _join_operands(
        self,
        other_operand: typing.Union[str, FeatureSet],
        join_type: str,
        asof_join: bool = False,
    ):
        if isinstance(other_operand, FeatureSet):
            other_operand = other_operand.metadata.name

        first_key_num = len(self._steps.keys()) if self._steps else 0
        left_last_step_name, left_all_feature_sets = (
            self.last_step_name,
            self.all_feature_sets_names,
        )
        is_first_fs = (
            join_type == JoinGraph.first_join_type or left_all_feature_sets == self.name
        )
        # create_new_step
        new_step = _JoinStep(
            f"step_{first_key_num}",
            left_last_step_name if not is_first_fs else "",
            other_operand,
            left_all_feature_sets if not is_first_fs else [],
            other_operand,
            join_type,
            asof_join,
        )

        if self.steps is not None:
            self.steps.update(new_step)
        else:
            self.steps = [new_step]
        return self

    def _start(self, other_operand: typing.Union[str, FeatureSet]):
        return self._join_operands(other_operand, JoinGraph.first_join_type)

    def _init_all_join_keys(
        self,
        feature_set_objects,
        vector,
        entity_rows_keys: typing.Optional[list[str]] = None,
    ):
        for step in self.steps:
            step.init_join_keys(feature_set_objects, vector, entity_rows_keys)

    @property
    def all_feature_sets_names(self):
        """
         Returns a list of all feature set names included in the join graph.

        :return:                    List[str]: A list of feature set names.
        """
        if self._steps:
            return self._steps[-1].left_feature_set_names + [
                self._steps[-1].right_feature_set_name
            ]
        else:
            return self.name

    @property
    def last_step_name(self):
        """
        Returns the name of the last step in the join graph.

        :return:                    str: The name of the last step.
        """
        if self._steps:
            return self._steps[-1].name
        else:
            return self.name

    @property
    def steps(self):
        """
        Returns the list of join steps as ObjectList, which can be used to iterate over the steps
        or access the properties of each step.
        :return:                    ObjectList: The list of join steps.
        """
        return self._steps

    @steps.setter
    def steps(self, steps):
        """
         Setter for the steps property. It allows updating the join steps.

        :param steps:               (List[_JoinStep]) The list of join steps.
        """
        self._steps = ObjectList.from_list(child_class=_JoinStep, children=steps)


class OnlineVectorService:
    """get_online_feature_service response object"""

    def __init__(
        self,
        vector,
        graph,
        index_columns,
        impute_policy: typing.Optional[dict] = None,
        requested_columns: typing.Optional[list[str]] = None,
    ):
        self.vector = vector
        self.impute_policy = impute_policy or {}

        self._controller = graph.controller
        self._index_columns = index_columns
        self._impute_values = {}
        self._requested_columns = requested_columns

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def initialize(self):
        """internal, init the feature service and prep the imputing logic"""
        if not self.impute_policy:
            return

        impute_policy = copy(self.impute_policy)
        vector = self.vector
        feature_stats = vector.get_stats_table()
        self._impute_values = {}

        feature_keys = list(vector.status.features.keys())
        if vector.status.label_column in feature_keys:
            feature_keys.remove(vector.status.label_column)

        if "*" in impute_policy:
            value = impute_policy["*"]
            del impute_policy["*"]

            for name in feature_keys:
                if name not in impute_policy:
                    if isinstance(value, str) and value.startswith("$"):
                        self._impute_values[name] = feature_stats.loc[name, value[1:]]
                    else:
                        self._impute_values[name] = value

        for name, value in impute_policy.items():
            if name not in feature_keys:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"feature {name} in impute_policy but not in feature vector"
                )
            if isinstance(value, str) and value.startswith("$"):
                self._impute_values[name] = feature_stats.loc[name, value[1:]]
            else:
                self._impute_values[name] = value

    @property
    def status(self):
        """vector merger function status (ready, running, error)"""
        return "ready"

    def get(self, entity_rows: list[Union[dict, list]], as_list=False):
        """get feature vector given the provided entity inputs

        take a list of input vectors/rows and return a list of enriched feature vectors
        each input and/or output vector can be a list of values or a dictionary of field names and values,
        to return the vector as a list of values set the `as_list` to True.

        if the input is a list of list (vs a list of dict), the values in the list will correspond to the
        index/entity values, i.e. [["GOOG"], ["MSFT"]] means "GOOG" and "MSFT" are the index/entity fields.

        example::

            # accept list of dict, return list of dict
            svc = fstore.get_online_feature_service(vector)
            resp = svc.get([{"name": "joe"}, {"name": "mike"}])

            # accept list of list, return list of list
            svc = fstore.get_online_feature_service(vector, as_list=True)
            resp = svc.get([["joe"], ["mike"]])

        :param entity_rows:  list of list/dict with input entity data/rows
        :param as_list:      return a list of list (list input is required by many ML frameworks)
        """
        results = []
        futures = []
        if isinstance(entity_rows, dict):
            entity_rows = [entity_rows]

        # validate we have valid input struct
        if (
            not entity_rows
            or not isinstance(entity_rows, list)
            or not isinstance(entity_rows[0], (list, dict))
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"input data is of type {type(entity_rows)}. must be a list of lists or list of dicts"
            )

        # if list of list, convert to dicts (with the index columns as the dict keys)
        if isinstance(entity_rows[0], list):
            if not self._index_columns or len(entity_rows[0]) != len(
                self._index_columns
            ):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "input list must be in the same size of the index_keys list"
                )
            index_range = range(len(self._index_columns))
            entity_rows = [
                {self._index_columns[i]: item[i] for i in index_range}
                for item in entity_rows
            ]

        for row in entity_rows:
            futures.append(self._controller.emit(row, return_awaitable_result=True))

        for future in futures:
            result = future.await_result()
            data = result.body
            if data:
                actual_columns = data.keys()
                if all([col in self._index_columns for col in actual_columns]):
                    # didn't get any data from the graph
                    results.append(None)
                    continue
                for column in self._requested_columns:
                    if (
                        column not in actual_columns
                        and column != self.vector.status.label_column
                    ):
                        data[column] = None

                if self._impute_values:
                    for name in data.keys():
                        v = data[name]
                        if v is None or (
                            isinstance(v, float) and (np.isinf(v) or np.isnan(v))
                        ):
                            data[name] = self._impute_values.get(name, v)
                if not self.vector.spec.with_indexes:
                    for name in self.vector.status.index_keys:
                        data.pop(name, None)
                if not any(data.values()):
                    data = None

            if as_list and data:
                data = [
                    data.get(key, None)
                    for key in self._requested_columns
                    if key != self.vector.status.label_column
                ]
            results.append(data)

        return results

    def close(self):
        """terminate the async loop"""
        self._controller.terminate()


class OfflineVectorResponse:
    """get_offline_features response object"""

    def __init__(self, merger):
        self._merger = merger
        self.vector = merger.vector

    @property
    def status(self):
        """vector prep job status (ready, running, error)"""
        return self._merger.get_status()

    def to_dataframe(self, to_pandas=True):
        """return result as dataframe"""
        if self.status != "completed":
            raise mlrun.errors.MLRunTaskNotReadyError(
                "feature vector dataset is not ready"
            )
        return self._merger.get_df(to_pandas=to_pandas)

    def to_parquet(self, target_path, **kw):
        """return results as parquet file"""
        return self._merger.to_parquet(target_path, **kw)

    def to_csv(self, target_path, **kw):
        """return results as csv file"""
        return self._merger.to_csv(target_path, **kw)



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mlrun


def get_snowflake_password():
    key = "SNOWFLAKE_PASSWORD"
    snowflake_password = mlrun.get_secret_or_env(key)

    if not snowflake_password:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"No password provided. Set password using the {key} "
            "project secret or environment variable."
        )

    return snowflake_password


def get_snowflake_spark_options(attributes):
    if not attributes:
        return {}
    return {
        "format": "net.snowflake.spark.snowflake",
        "sfURL": attributes.get("url"),
        "sfUser": attributes.get("user"),
        "sfPassword": get_snowflake_password(),
        "sfDatabase": attributes.get("database"),
        "sfSchema": attributes.get("db_schema"),
        "sfWarehouse": attributes.get("warehouse"),
        "application": "iguazio_platform",
        "TIMESTAMP_TYPE_MAPPING": "TIMESTAMP_LTZ",
    }



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "DataItem",
    "ModelProvider",
    "get_store_resource",
    "ParquetTarget",
    "CSVTarget",
    "NoSqlTarget",
    "StreamTarget",
    "KafkaTarget",
    "CSVSource",
    "ParquetSource",
    "BigQuerySource",
    "HttpSource",
    "StreamSource",
    "KafkaSource",
    "RedisStore",
    "DatabricksFileSystemDisableCache",
    "DatabricksFileBugFixed",
    "get_stream_pusher",
    "ConfigProfile",
    "VectorStoreCollection",
    "store_manager",
]

from urllib.parse import urlparse

import fsspec
import storey

import mlrun.datastore.wasbfs
from mlrun.datastore.datastore_profile import (
    DatastoreProfileKafkaStream,
    DatastoreProfileKafkaTarget,
    DatastoreProfileV3io,
)
from mlrun.datastore.model_provider.model_provider import ModelProvider
from mlrun.platforms.iguazio import (
    HTTPOutputStream,
    KafkaOutputStream,
    OutputStream,
    parse_path,
)

from ..utils import logger
from .base import DataItem
from .datastore import StoreManager, in_memory_store, uri_to_ipython
from .dbfs_store import DatabricksFileBugFixed, DatabricksFileSystemDisableCache
from .s3 import parse_s3_bucket_and_key
from .sources import (
    BigQuerySource,
    CSVSource,
    HttpSource,
    KafkaSource,
    ParquetSource,
    StreamSource,
)
from .store_resources import (
    get_store_resource,
    get_store_uri,
    is_store_uri,
    parse_store_uri,
)
from .targets import CSVTarget, NoSqlTarget, ParquetTarget, StreamTarget
from .utils import get_kafka_brokers_from_dict, parse_kafka_url

store_manager = StoreManager()

if hasattr(fsspec, "register_implementation"):
    fsspec.register_implementation(
        "dbfs", DatabricksFileSystemDisableCache, clobber=True
    )
else:
    from fsspec.registry import known_implementations

    known_implementations["dbfs"] = {
        "class": "mlrun.datastore.dbfs_store.DatabricksFileSystemDisableCache",
        "err": "Please make sure your fsspec version supports dbfs",
    }

    del known_implementations

del fsspec  # clear the module namespace


def set_in_memory_item(key, value):
    item = store_manager.object(f"memory://{key}")
    item.put(value)
    return item


def get_in_memory_items():
    return in_memory_store._items


def get_stream_pusher(stream_path: str, **kwargs):
    """get a stream pusher object from URL.

    common kwargs::

        create:             create a new stream if doesnt exist
        shards:             number of shards
        retention_in_hours: stream retention in hours

    :param stream_path:        path/url of stream
    """
    if stream_path.startswith("ds://"):
        datastore_profile = mlrun.datastore.datastore_profile.datastore_profile_read(
            stream_path
        )
        if isinstance(
            datastore_profile,
            (DatastoreProfileKafkaStream, DatastoreProfileKafkaTarget),
        ):
            attributes = datastore_profile.attributes()
            brokers = attributes.pop("brokers", None)
            # Override the topic with the one in the url (if any)
            parsed_url = urlparse(stream_path)
            topic = (
                parsed_url.path.strip("/")
                if parsed_url.path
                else datastore_profile.get_topic()
            )
            producer_options = mlrun.datastore.utils.KafkaParameters(
                attributes
            ).producer()
            return KafkaOutputStream(topic, brokers, producer_options=producer_options)

        elif isinstance(datastore_profile, DatastoreProfileV3io):
            parsed_url = urlparse(stream_path)
            stream_path = datastore_profile.url(parsed_url.path)
            endpoint, stream_path = parse_path(stream_path)
            return OutputStream(stream_path, endpoint=endpoint, **kwargs)
        else:
            raise ValueError(
                f"Unsupported datastore profile type: {type(datastore_profile)}"
            )
    else:
        kafka_brokers = get_kafka_brokers_from_dict(kwargs)
        if stream_path.startswith("kafka://") or kafka_brokers:
            topic, brokers = parse_kafka_url(stream_path, kafka_brokers)
            return KafkaOutputStream(
                topic, brokers, kwargs.get("kafka_producer_options")
            )
        elif stream_path.startswith("http://") or stream_path.startswith("https://"):
            return HTTPOutputStream(stream_path=stream_path)
        elif "://" not in stream_path:
            return OutputStream(stream_path, **kwargs)
        elif stream_path.startswith("v3io"):
            endpoint, stream_path = parse_path(stream_path)
            endpoint = kwargs.pop("endpoint", None) or endpoint
            return OutputStream(stream_path, endpoint=endpoint, **kwargs)
        elif stream_path.startswith("dummy://"):
            return _DummyStream(**kwargs)
        else:
            raise ValueError(f"unsupported stream path {stream_path}")


class _DummyStream(storey.MapClass):
    """stream emulator for tests and debug"""

    def __init__(self, event_list=None, **kwargs):
        self.event_list = event_list or []
        super().__init__(**kwargs)

    def push(self, data, **kwargs):
        if not isinstance(data, list):
            data = [data]
        for item in data:
            logger.info(f"dummy stream got event: {item}, kwargs={kwargs}")
            self.event_list.append(item)

    def do(self, event):
        if not isinstance(event, list):
            event = [event]
        for item in event:
            self.event_list.append(item)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from typing import Optional

import cloudpickle

import mlrun

from .._common import without_mlrun_interface
from .._ml_common import MLModelHandler
from .mlrun_interface import SKLearnMLRunInterface
from .utils import SKLearnTypes


class SKLearnModelHandler(MLModelHandler):
    """
    Class for handling a SciKitLearn model, enabling loading and saving it during runs.
    """

    # Framework name:
    FRAMEWORK_NAME = "sklearn"

    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.

        :raise MLRunNotFoundError: If the model file was not found.
        """
        # Get the pickle model file:
        self._model_file = os.path.join(self._model_path, f"{self._model_name}.pkl")
        if not os.path.exists(self._model_file):
            raise mlrun.errors.MLRunNotFoundError(
                f"The model file '{self._model_name}.pkl' was not found within the given 'model_path': "
                f"'{self._model_path}'"
            )

    @without_mlrun_interface(interface=SKLearnMLRunInterface)
    def save(self, output_path: Optional[str] = None, **kwargs):
        """
        Save the handled model at the given output path. If a MLRun context is available, the saved model files will be
        logged and returned as artifacts.

        :param output_path: The full path to the directory to save the handled model at. If not given, the context
                            stored will be used to save the model in the default artifacts location.

        :return The saved model additional artifacts (if needed) dictionary if context is available and None otherwise.
        """
        super().save(output_path=output_path)

        # Save the model pkl file:
        self._model_file = f"{self._model_name}.pkl"
        with open(self._model_file, "wb") as pickle_file:
            cloudpickle.dump(self._model, pickle_file)

        return None

    def load(self, **kwargs):
        """
        Load the specified model in this handler. Additional parameters for the class initializer can be passed via the
        kwargs dictionary.
        """
        super().load()

        # Load from a pkl file:
        with open(self._model_file, "rb") as pickle_file:
            self._model = pickle.load(pickle_file)

    def to_onnx(
        self,
        model_name: Optional[str] = None,
        optimize: bool = True,
        input_sample: SKLearnTypes.DatasetType = None,
        log: Optional[bool] = None,
    ):
        """
        Convert the model in this handler to an ONNX model. The inputs names are optional, they do not change the
        semantics of the model, it is only for readability.

        :param model_name:          The name to give to the converted ONNX model. If not given the default name will be
                                    the stored model name with the suffix '_onnx'.
        :param optimize:            Whether to optimize the ONNX model using 'onnxoptimizer' before saving the model.
                                    Default: True.
        :param input_sample:        An inputs sample with the names and data types of the inputs of the model.
        :param log:                 In order to log the ONNX model, pass True. If None, the model will be logged if this
                                    handler has a MLRun context set. Default: None.

        :return: The converted ONNX model (onnx.ModelProto).

        :raise MLRunMissingDependencyError: If some of the ONNX packages are missing.
        """
        # Import onnx related modules:
        try:
            pass
            # import skl2onnx

            # from mlrun.frameworks.onnx import ONNXModelHandler
        except ModuleNotFoundError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "ONNX conversion requires additional packages to be installed. "
                "Please run 'pip install mlrun[sklearn]' to install MLRun's XGBoost package."
            )

        raise NotImplementedError  # TODO: Finish ONNX conversion



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

from .function import Function
from .k8s import Resources


class ClientSpec(pydantic.v1.BaseModel):
    version: typing.Optional[str]
    namespace: typing.Optional[str]
    docker_registry: typing.Optional[str]
    remote_host: typing.Optional[str]
    mpijob_crd_version: typing.Optional[str]
    ui_url: typing.Optional[str]
    artifact_path: typing.Optional[str]
    feature_store_data_prefixes: typing.Optional[dict[str, str]]
    feature_store_default_targets: typing.Optional[str]
    spark_app_image: typing.Optional[str]
    spark_app_image_tag: typing.Optional[str]
    spark_history_server_path: typing.Optional[str]
    spark_operator_version: typing.Optional[str]
    kfp_image: typing.Optional[str]
    kfp_url: typing.Optional[str]
    dask_kfp_image: typing.Optional[str]
    api_url: typing.Optional[str]
    nuclio_version: typing.Optional[str]
    ui_projects_prefix: typing.Optional[str]
    scrape_metrics: typing.Optional[str]
    default_function_node_selector: typing.Optional[str]
    igz_version: typing.Optional[str]
    auto_mount_type: typing.Optional[str]
    auto_mount_params: typing.Optional[str]
    default_function_priority_class_name: typing.Optional[str]
    valid_function_priority_class_names: typing.Optional[str]
    default_tensorboard_logs_path: typing.Optional[str]
    default_function_pod_resources: typing.Optional[Resources]
    preemptible_nodes_node_selector: typing.Optional[str]
    preemptible_nodes_tolerations: typing.Optional[str]
    default_preemption_mode: typing.Optional[str]
    force_run_local: typing.Optional[str]
    function: typing.Optional[Function]
    redis_url: typing.Optional[str]
    redis_type: typing.Optional[str]
    sql_url: typing.Optional[str]
    ce: typing.Optional[dict]
    # not passing them as one object as it possible client user would like to override only one of the params
    calculate_artifact_hash: typing.Optional[str]
    generate_artifact_target_path_from_artifact_hash: typing.Optional[str]
    logs: typing.Optional[dict]
    packagers: typing.Optional[dict]
    external_platform_tracking: typing.Optional[dict]
    alerts_mode: typing.Optional[str]
    system_id: typing.Optional[str]
    model_endpoint_monitoring_store_prefixes: typing.Optional[dict[str, str]]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from lightgbm.callback import CallbackEnv


class Callback(ABC):
    """
    An abstract callback class for writing callbacks usable by LightGBM's training. The function expects a callable
    object, as such the `__call__` method will be used.

    There are two configurable class properties:

    * order: int = 10 - The priority of the callback to be called first. Lower value means higher priority. Default:
      10.
    * before_iteration: bool = False - Whether to call this callback before each iteration or after. Default: after
      (False).

    LightGBM will pass a `CallbackEnv` object to the callbacks - a `NamedTuple` with the following items:

    * model: Booster - The model's Booster object.
    * params: Dict - The training parameters passed to 'lgb.train'.
    * iteration: int - The current iteration.
    * begin_iteration: int - The first iteration.
    * end_iteration: int - The last iteration.
    * evaluation_result_list: List[Tuple[str, str, float, bool, Optional[float]]] - A list of metric tuples. Each metric
      tuple is constructed by:

      * [0] str - The validation set name the metric was calculated on.
      * [1] str - The metric name.
      * [2] float - The metric score (mean score in case of `lightgbm.cv`).
      * [3] bool - The aim of the metric, True means bigger score is better and False means smaller score is better.
      * [4] Optional[float] - The metric stdv score (only in case of `lightgbm.cv`)

    In addition, if the LightGBM module or model are wrapped with MLRun, the methods `on_train_begin` and `on_train_end`
    will be called as well.

    example::

        class ExampleCallback(Callback):
            def __init__(self, name: str):
                self.name = name

            def __call__(self, env: CallbackEnv):
                print(f"{self.name}: current iteration: {env.iteration}")

            def on_train_begin(self):
                print("{self.name}: Starting training...")

            def on_train_end(self):
                print("{self.name}: Done training!")


        apply_mlrun()
        lgb.train(..., callbacks=[ExampleCallback(name="Example")])
    """

    def __init__(self, order: int = 10, before_iteration: bool = False):
        """
        Initialize a new callback to use in LightGBM's training.

        :param order:            The priority of the callback to be called first. Lower value means higher priority.
                                 Default: 10.
        :param before_iteration: Whether to call this callback before each iteration or after. Default: after
                                 (False).
        """
        self.order = order
        self.before_iteration = before_iteration

    def __call__(self, env: CallbackEnv):
        """
        The method to be called during training of a LightGBM model. It will be called at the end of each iteration
        post validating on all the given validation datasets.

        :param env: The CallbackEnv representing the current iteration of the training.
        """
        pass

    def on_train_begin(self):
        """
        Method to be called before the training starts. Will only be called if the model is wrapped with an MLRun
        interface.
        """
        pass

    def on_train_end(self):
        """
        Method to be called after the training ends. Will only be called if the model is wrapped with an MLRun
        interface.
        """
        pass



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

import mlrun
from mlrun.artifacts import Artifact
from mlrun.utils.helpers import is_jupyter


class Plan(ABC):
    """
    An abstract class for describing a plan. A plan is used to produce artifact manually or in a given time of a
    function according to its configuration.
    """

    def __init__(self):
        """
        Initialize a new plan.
        """
        self._artifacts = {}  # type: Dict[str, Artifact]

    @property
    def artifacts(self) -> dict[str, Artifact]:
        """
        Get the plan's produced artifacts.

        :return: The plan's artifacts.
        """
        return self._artifacts

    def is_reproducible(self, *args, **kwargs) -> bool:
        """
        Check whether the plan should be used to produce multiple times or only once. Default: return False.

        :return: True if the plan is reproducible and False otherwise.
        """
        return False

    @abstractmethod
    def is_ready(self, *args, **kwargs) -> bool:
        """
        Check whether the plan is fit for production in the current time this method is called.

        :return: True if the plan is producible and False otherwise.
        """
        pass

    @abstractmethod
    def produce(self, *args, **kwargs) -> dict[str, Artifact]:
        """
        Produce the artifact according to this plan.

        :return: The produced artifacts.
        """
        pass

    def log(self, context: mlrun.MLClientCtx):
        """
        Log the artifacts in this plan to the given context.

        :param context: A MLRun context to log with.
        """
        for artifact_name, artifact_object in self._artifacts.items():
            context.log_artifact(artifact_object)

    def display(self):
        """
        Display the plan's artifact. If artifacts were not produced nothing will be printed.
        """
        # Validate at least one artifact was produced:
        if not self._artifacts:
            return

        # Call the correct display method according to the kernel:
        if is_jupyter:
            self._gui_display()
        else:
            self._cli_display()

    @abstractmethod
    def _cli_display(self):
        """
        How the plan's products would be presented on a command line kernel.
        """
        pass

    @abstractmethod
    def _gui_display(self):
        """
        How the plan's products would be presented on a graphic IPython kernel (like a Jupyter notebook).
        """
        pass

    def _repr_pretty_(self, p, cycle: bool):
        """
        A pretty representation of the plan. Will be called by the IPython kernel. This method will call the plan's
        display method.

        :param p:     A RepresentationPrinter instance.
        :param cycle: If a cycle is detected to prevent infinite loop.
        """
        self.display()
